# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict, List
import json
import unittest
from datetime import datetime, timedelta, date

import azure.functions as func
import azure.functions.servicebus as azf_sb
from azure.functions import meta

from tests.utils.testutils import (CollectionBytes, CollectionString,
                                   CollectionSint64)


class TestServiceBus(unittest.TestCase):
    MOCKED_CONTENT_TYPE = 'application/json'
    MOCKED_CORROLATION_ID = '87c66eaf88e84119b66a26278a7b4149'
    MOCKED_DEADLETTER_ERROR_DESCRIPTION = \
        'mocked_dead_letter_error_description'
    MOCKED_DEADLETTER_REASON = 'mocked_dead_letter_reason'
    MOCKED_DEADLETTER_SOURCE = 'mocked_dead_letter_source'
    MOCKED_DELIVERY_COUNT = 571
    MOCKED_ENQUEUED_SEQUENCE_NUMBER = 15
    MOCKED_ENQUEUE_TIME_UTC = datetime.utcnow()
    MOCKED_EXPIRY_AT_UTC = datetime.utcnow()
    MOCKED_LABEL = 'mocked_label'
    MOCKED_LOCKED_UNTIL = datetime.utcnow()
    MOCKED_LOCK_TOKEN = '87931fd2-39f4-415a-9fdc-adfdcbed3148'
    MOCKED_MESSAGE_ID = 'abcee18397398d93891830a0aac89eed'
    MOCKED_MESSAGE_ID_A = 'aaaaa18397398d93891830a0aac89eed'
    MOCKED_MESSAGE_ID_B = 'bbbbb18397398d93891830a0aac89eed'
    MOCKED_MESSAGE_ID_C = 'ccccc18397398d93891830a0aac89eed'
    MOCKED_PARTITION_KEY = 'mocked_partition_key'
    MOCKED_REPLY_TO = 'mocked_reply_to'
    MOCKED_REPLY_TO_SESSION_ID = 'mocked_reply_to_session_id'
    MOCKED_SCHEDULED_ENQUEUE_TIME_UTC = datetime.utcnow()
    MOCKED_SEQUENCE_NUMBER = 38291
    MOCKED_SESSION_ID = 'mocked_session_id'
    MOCKED_STATE = 1
    MOCKED_SUBJECT = 'mocked_subject'
    MOCKED_TIME_TO_LIVE = '11:22:33'
    MOCKED_TIME_TO_LIVE_TIMEDELTA = timedelta(hours=11, minutes=22, seconds=33)
    MOCKED_TO = 'mocked_to'
    MOCKED_TRANSACTION_PARTITION_KEY = 'mocked_transaction_partition_key'

    MOCKED_AZURE_PARTNER_ID = '6ceef68b-0794-45dd-bb2e-630748515552'

    def test_servicebusmessage_initialize_without_args(self):
        # given
        expected_body = b""
        expexceted_content_type = None
        expected_correlation_id = None

        # when
        test_sb_message = func.ServiceBusMessage()

        # then
        assert expected_body == test_sb_message.get_body()
        assert expexceted_content_type == test_sb_message.content_type
        assert expected_correlation_id == test_sb_message.correlation_id

    def test_servicebusmessage_initialize_all_arguments(self):
        # given
        expected_body: bytes = b"Body"
        expected_content_type: str = "Content Type"
        expected_correlation_id: str = "Correlation ID"

        # when
        test_sb_message = func.ServiceBusMessage(
            body=expected_body,
            content_type=expected_content_type,
            correlation_id=expected_correlation_id
        )

        # then
        assert expected_body == test_sb_message.get_body()
        assert expected_content_type == test_sb_message.content_type
        assert expected_correlation_id == test_sb_message.correlation_id
        self.assertDictEqual(test_sb_message.application_properties, {})
        self.assertIsNone(test_sb_message.dead_letter_error_description)
        self.assertIsNone(test_sb_message.dead_letter_reason)
        self.assertIsNone(test_sb_message.dead_letter_source)
        self.assertIsNone(test_sb_message.delivery_count)
        self.assertIsNone(test_sb_message.enqueued_sequence_number)
        self.assertIsNone(test_sb_message.enqueued_time_utc)
        self.assertIsNone(test_sb_message.expires_at_utc)
        self.assertIsNone(test_sb_message.expiration_time)
        self.assertIsNone(test_sb_message.label)
        self.assertIsNone(test_sb_message.locked_until)
        self.assertIsNone(test_sb_message.lock_token)
        assert "" == test_sb_message.message_id
        self.assertIsNone(test_sb_message.partition_key)
        self.assertIsNone(test_sb_message.reply_to)
        self.assertIsNone(test_sb_message.reply_to_session_id)
        self.assertIsNone(test_sb_message.scheduled_enqueue_time)
        self.assertIsNone(test_sb_message.scheduled_enqueue_time_utc)
        self.assertIsNone(test_sb_message.sequence_number)
        self.assertIsNone(test_sb_message.session_id)
        self.assertIsNone(test_sb_message.state)
        self.assertIsNone(test_sb_message.subject)
        self.assertIsNone(test_sb_message.time_to_live)
        self.assertIsNone(test_sb_message.to)
        self.assertIsNone(test_sb_message.transaction_partition_key)
        self.assertDictEqual(test_sb_message.user_properties, {})
        self.assertIsNone(test_sb_message.metadata)

    def test_servicebus_message_initialize_all_args(self):
        # given
        body = "body"
        trigger_metadata = "trigger metadata"
        application_properties = {"application": "properties"}
        content_type = "content type"
        correlation_id = "correlation id"
        dead_letter_error_description = "dead letter error description"
        dead_letter_reason = "dead letter reason"
        dead_letter_source = "dead letter source"
        delivery_count = 1
        enqueued_sequence_number = 1
        enqueued_time_utc = date(2022, 5, 1)
        expires_at_utc = date(2022, 5, 1)
        label = "label"
        locked_until = date(2022, 5, 1)
        lock_token = "lock token"
        message_id = "message id"
        partition_key = "partition key"
        reply_to = "reply to"
        reply_to_session_id = "reply to session id"
        scheduled_enqueue_time_utc = date(2022, 5, 1)
        sequence_number = 1
        session_id = "session id"
        state = 1
        subject = "subject"
        time_to_live = timedelta(hours=1)
        to = "to"
        transaction_partition_key = "transaction partition key"
        user_properties = {"user": "properties"}

        # when
        sb_message = azf_sb.ServiceBusMessage(
            body=body,
            trigger_metadata=trigger_metadata,
            application_properties=application_properties,
            content_type=content_type,
            correlation_id=correlation_id,
            dead_letter_error_description=dead_letter_error_description,
            dead_letter_reason=dead_letter_reason,
            dead_letter_source=dead_letter_source,
            delivery_count=delivery_count,
            enqueued_sequence_number=enqueued_sequence_number,
            enqueued_time_utc=enqueued_time_utc,
            expires_at_utc=expires_at_utc,
            label=label,
            locked_until=locked_until,
            lock_token=lock_token,
            message_id=message_id,
            partition_key=partition_key,
            reply_to=reply_to,
            reply_to_session_id=reply_to_session_id,
            scheduled_enqueue_time_utc=scheduled_enqueue_time_utc,
            sequence_number=sequence_number,
            session_id=session_id,
            state=state,
            subject=subject,
            time_to_live=time_to_live,
            to=to,
            transaction_partition_key=transaction_partition_key,
            user_properties=user_properties)

        # then
        self.assertEqual(sb_message.get_body(), body)
        self.assertEqual(sb_message.application_properties,
                         application_properties)
        self.assertEqual(sb_message.content_type, content_type)
        self.assertEqual(sb_message.correlation_id, correlation_id)
        self.assertEqual(sb_message.dead_letter_error_description,
                         dead_letter_error_description)
        self.assertEqual(sb_message.dead_letter_reason, dead_letter_reason)
        self.assertEqual(sb_message.dead_letter_source, dead_letter_source)
        self.assertEqual(sb_message.delivery_count, delivery_count)
        self.assertEqual(sb_message.enqueued_sequence_number,
                         enqueued_sequence_number)
        self.assertEqual(sb_message.enqueued_time_utc, enqueued_time_utc)
        self.assertEqual(sb_message.expires_at_utc, expires_at_utc)
        self.assertEqual(sb_message.label, label)
        self.assertEqual(sb_message.locked_until, locked_until)
        self.assertEqual(sb_message.lock_token, lock_token)
        self.assertEqual(sb_message.message_id, message_id)
        self.assertEqual(sb_message.partition_key, partition_key)
        self.assertEqual(sb_message.reply_to, reply_to)
        self.assertEqual(sb_message.reply_to_session_id, reply_to_session_id)
        self.assertEqual(sb_message.scheduled_enqueue_time_utc,
                         scheduled_enqueue_time_utc)
        self.assertEqual(sb_message.sequence_number, sequence_number)
        self.assertEqual(sb_message.session_id, session_id)
        self.assertEqual(sb_message.state, state)
        self.assertEqual(sb_message.subject, subject)
        self.assertEqual(sb_message.time_to_live, time_to_live)
        self.assertEqual(sb_message.to, to)
        self.assertEqual(sb_message.transaction_partition_key,
                         transaction_partition_key)
        self.assertEqual(sb_message.user_properties, user_properties)

    def test_abstract_servicebus_message(self):
        test_sb_message = func.ServiceBusMessage()
        abstract_sb_message = func._abc.ServiceBusMessage

        self.assertIsInstance(test_sb_message, abstract_sb_message)
        with self.assertRaises(TypeError):
            func._abc.ServiceBusMessage()

    def test_servicebus_input_type(self):
        check_input_type = (
            azf_sb.ServiceBusMessageInConverter.check_input_type_annotation
        )
        # Should accept a single service bus message as trigger input type
        self.assertTrue(check_input_type(func.ServiceBusMessage))

        # Should accept a multiple service bus message as trigger input type
        self.assertTrue(check_input_type(List[func.ServiceBusMessage]))

        # Should accept a message class derived from service bus
        class ServiceBusMessageChild(func.ServiceBusMessage):
            FOO = 'BAR'

        self.assertTrue(check_input_type(ServiceBusMessageChild))

        # Should be false if a message type does not match expectation
        self.assertFalse(check_input_type(func.EventHubEvent))
        self.assertFalse(check_input_type(str))
        self.assertFalse(check_input_type(type(None)))

    def test_servicebus_output_type(self):
        check_output_type = (
            azf_sb.ServiceBusMessageOutConverter.check_output_type_annotation
        )
        # Should accept bytes and string as trigger output type
        self.assertTrue(check_output_type(bytes))
        self.assertTrue(check_output_type(str))

        # Should reject if attempt to send a service bus message out
        self.assertFalse(check_output_type(func.ServiceBusMessage))

        # Should be false if a message type does not match expectation
        self.assertFalse(check_output_type(func.EventGridEvent))
        self.assertFalse(check_output_type(type(None)))

    def test_servicebus_data(self):
        servicebus_msg = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_single_servicebus_data(),
            trigger_metadata=self._generate_single_trigger_metadata())

        servicebus_data = servicebus_msg.get_body().decode('utf-8')
        self.assertEqual(servicebus_data, json.dumps({"lucky_number": 23}))

    def test_servicebus_non_existing_property(self):
        # The function should not fail even when property does not work
        msg = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_single_servicebus_data(),
            trigger_metadata={
                'MessageId': meta.Datum(self.MOCKED_MESSAGE_ID, 'string'),
                'UserProperties': meta.Datum('{ "UserId": 1 }', 'json')
            }
        )

        # Property that are not passed from extension should be None
        self.assertIsNone(msg.content_type)

        # Message id should always be available
        self.assertEqual(msg.message_id, self.MOCKED_MESSAGE_ID)

        # User property should always be available
        self.assertEqual(msg.user_properties['UserId'], 1)

    def test_servicebus_properties(self):
        # SystemProperties in metadata should propagate to class properties
        msg = azf_sb.ServiceBusMessageInConverter.decode(
            data=meta.Datum(b'body_bytes', 'bytes'),
            trigger_metadata=self._generate_single_trigger_metadata())

        self.assertEqual(msg.get_body(), b'body_bytes')

        # Test individual ServiceBus properties respectively
        self.assertEqual(msg.application_properties,
                         {'application': 'value'})
        self.assertEqual(msg.content_type,
                         self.MOCKED_CONTENT_TYPE)
        self.assertEqual(msg.correlation_id,
                         self.MOCKED_CORROLATION_ID)
        self.assertEqual(msg.dead_letter_error_description,
                         self.MOCKED_DEADLETTER_ERROR_DESCRIPTION)
        self.assertEqual(msg.dead_letter_reason, self.MOCKED_DEADLETTER_REASON)
        self.assertEqual(msg.dead_letter_source,
                         self.MOCKED_DEADLETTER_SOURCE)
        self.assertEqual(msg.enqueued_sequence_number,
                         self.MOCKED_ENQUEUED_SEQUENCE_NUMBER)
        self.assertEqual(msg.enqueued_time_utc,
                         self.MOCKED_ENQUEUE_TIME_UTC)
        self.assertEqual(msg.expires_at_utc,
                         self.MOCKED_EXPIRY_AT_UTC)
        self.assertEqual(msg.expiration_time,
                         self.MOCKED_EXPIRY_AT_UTC)
        self.assertEqual(msg.label,
                         self.MOCKED_LABEL)
        self.assertEqual(msg.locked_until,
                         self.MOCKED_LOCKED_UNTIL)
        self.assertEqual(msg.lock_token,
                         self.MOCKED_LOCK_TOKEN)
        self.assertEqual(msg.message_id,
                         self.MOCKED_MESSAGE_ID)
        self.assertEqual(msg.partition_key,
                         self.MOCKED_PARTITION_KEY)
        self.assertEqual(msg.reply_to,
                         self.MOCKED_REPLY_TO)
        self.assertEqual(msg.reply_to_session_id,
                         self.MOCKED_REPLY_TO_SESSION_ID)
        self.assertEqual(msg.scheduled_enqueue_time,
                         self.MOCKED_SCHEDULED_ENQUEUE_TIME_UTC)
        self.assertEqual(msg.scheduled_enqueue_time_utc,
                         self.MOCKED_SCHEDULED_ENQUEUE_TIME_UTC)
        self.assertEqual(msg.session_id,
                         self.MOCKED_SESSION_ID)
        self.assertEqual(msg.state,
                         self.MOCKED_STATE)
        self.assertEqual(msg.subject,
                         self.MOCKED_SUBJECT)
        self.assertEqual(msg.time_to_live,
                         self.MOCKED_TIME_TO_LIVE_TIMEDELTA)
        self.assertEqual(msg.to,
                         self.MOCKED_TO)
        self.assertEqual(msg.transaction_partition_key,
                         self.MOCKED_TRANSACTION_PARTITION_KEY)
        self.assertDictEqual(msg.user_properties, {
            '$AzureWebJobsParentId': self.MOCKED_AZURE_PARTNER_ID,
            'x-opt-enqueue-sequence-number': 0
        })

    def test_servicebus_metadata(self):
        # Trigger metadata should contains all the essential information
        # about this service bus message
        servicebus_msg = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_single_servicebus_data(),
            trigger_metadata=self._generate_single_trigger_metadata())

        # Datetime should be in iso8601 string instead of datetime object
        metadata_dict = servicebus_msg.metadata
        self.assertGreaterEqual(metadata_dict.items(), {
            'DeliveryCount': self.MOCKED_DELIVERY_COUNT,
            'LockToken': self.MOCKED_LOCK_TOKEN,
            'ExpiresAtUtc': self.MOCKED_EXPIRY_AT_UTC.isoformat(),
            'EnqueuedTimeUtc': self.MOCKED_ENQUEUE_TIME_UTC.isoformat(),
            'MessageId': self.MOCKED_MESSAGE_ID,
            'ContentType': self.MOCKED_CONTENT_TYPE,
            'SequenceNumber': self.MOCKED_SEQUENCE_NUMBER,
            'Label': self.MOCKED_LABEL,
            'sys': {
                'MethodName': 'ServiceBusSMany',
                'UtcNow': '2020-06-18T05:39:12.2860411Z',
                'RandGuid': 'bb38deae-cc75-49f2-89f5-96ec6eb857db'
            }
        }.items())

    def test_servicebus_should_not_override_metadata(self):
        # SystemProperties in metadata should propagate to class properties
        servicebus_msg = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_single_servicebus_data(),
            trigger_metadata=self._generate_single_trigger_metadata())

        # The content_type trigger field should be set
        self.assertEqual(servicebus_msg.content_type, 'application/json')

        # The metadata field should also be set
        self.assertEqual(servicebus_msg.metadata['ContentType'],
                         'application/json')

        # Now we change the metadata field
        # The trigger property should still remain the same
        servicebus_msg.metadata['ContentType'] = 'text/plain'
        self.assertEqual(servicebus_msg.content_type, 'application/json')

    def test_multiple_servicebus_trigger(self):
        # When cardinality is turned on to 'many', metadata should contain
        # information for all messages
        servicebus_msgs = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_multiple_service_bus_data(),
            trigger_metadata=self._generate_multiple_trigger_metadata()
        )

        # The decoding result should contain a list of message
        self.assertEqual(len(servicebus_msgs), 3)

    def test_multiple_servicebus_trigger_non_existing_properties(self):
        servicebus_msgs = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_multiple_service_bus_data(),
            trigger_metadata={
                'MessageIdArray': meta.Datum(type='collection_string',
                                             value=CollectionString([
                                                 self.MOCKED_MESSAGE_ID,
                                                 self.MOCKED_MESSAGE_ID,
                                                 self.MOCKED_MESSAGE_ID
                                             ])),
                'UserPropertiesArray': meta.Datum(type='json',
                                                  value='''[{ "UserId": 1 },
                                                            { "UserId": 2 },
                                                            { "UserId": 3 }]
                                                        ''')
            }
        )

        # Non existing properties should return None
        self.assertIsNone(servicebus_msgs[0].content_type)
        self.assertIsNone(servicebus_msgs[1].content_type)
        self.assertIsNone(servicebus_msgs[2].content_type)

        # Message Id should always be available
        self.assertEqual(servicebus_msgs[0].message_id, self.MOCKED_MESSAGE_ID)
        self.assertEqual(servicebus_msgs[1].message_id, self.MOCKED_MESSAGE_ID)
        self.assertEqual(servicebus_msgs[2].message_id, self.MOCKED_MESSAGE_ID)

        # User properties should always be available
        self.assertEqual(servicebus_msgs[0].user_properties['UserId'], 1)
        self.assertEqual(servicebus_msgs[1].user_properties['UserId'], 2)
        self.assertEqual(servicebus_msgs[2].user_properties['UserId'], 3)

    def test_multiple_servicebus_trigger_properties(self):
        # When cardinality is turned on to 'many', metadata should contain
        # information for all messages
        servicebus_msgs = azf_sb.ServiceBusMessageInConverter.decode(
            data=self._generate_multiple_service_bus_data(),
            trigger_metadata=self._generate_multiple_trigger_metadata()
        )

        expected_bodies: List[str] = [
            json.dumps({"lucky_number": 23}),
            json.dumps({"lucky_number": 34}),
            json.dumps({"lucky_number": 45}),
        ]

        expected_message_ids: List[int] = [
            self.MOCKED_MESSAGE_ID_A,
            self.MOCKED_MESSAGE_ID_B,
            self.MOCKED_MESSAGE_ID_C
        ]

        for i in range(len(servicebus_msgs)):
            msg = servicebus_msgs[i]
            body_data = msg.get_body().decode('utf-8')
            self.assertEqual(body_data, expected_bodies[i])
            self.assertDictEqual(msg.application_properties,
                                 {"application": "value"})
            self.assertEqual(msg.content_type,
                             self.MOCKED_CONTENT_TYPE)
            self.assertEqual(msg.correlation_id,
                             self.MOCKED_CORROLATION_ID)
            self.assertEqual(msg.dead_letter_error_description,
                             self.MOCKED_DEADLETTER_ERROR_DESCRIPTION)
            self.assertEqual(msg.dead_letter_reason,
                             self.MOCKED_DEADLETTER_REASON)
            self.assertEqual(msg.dead_letter_source,
                             self.MOCKED_DEADLETTER_SOURCE)
            self.assertEqual(msg.enqueued_sequence_number,
                             self.MOCKED_ENQUEUED_SEQUENCE_NUMBER)
            self.assertEqual(msg.enqueued_time_utc,
                             self.MOCKED_ENQUEUE_TIME_UTC)
            self.assertEqual(msg.expires_at_utc,
                             self.MOCKED_EXPIRY_AT_UTC)
            self.assertEqual(msg.expiration_time,
                             self.MOCKED_EXPIRY_AT_UTC)
            self.assertEqual(msg.label,
                             self.MOCKED_LABEL)
            self.assertEqual(msg.locked_until,
                             self.MOCKED_LOCKED_UNTIL)
            self.assertEqual(msg.lock_token,
                             self.MOCKED_LOCK_TOKEN)
            self.assertEqual(msg.message_id,
                             expected_message_ids[i])
            self.assertEqual(msg.partition_key,
                             self.MOCKED_PARTITION_KEY)
            self.assertEqual(msg.reply_to,
                             self.MOCKED_REPLY_TO)
            self.assertEqual(msg.reply_to_session_id,
                             self.MOCKED_REPLY_TO_SESSION_ID)
            self.assertEqual(msg.scheduled_enqueue_time,
                             self.MOCKED_SCHEDULED_ENQUEUE_TIME_UTC)
            self.assertEqual(msg.scheduled_enqueue_time_utc,
                             self.MOCKED_SCHEDULED_ENQUEUE_TIME_UTC)
            self.assertEqual(msg.session_id,
                             self.MOCKED_SESSION_ID)
            self.assertEqual(msg.state,
                             self.MOCKED_STATE)
            self.assertEqual(msg.subject,
                             self.MOCKED_SUBJECT)
            self.assertEqual(msg.time_to_live,
                             self.MOCKED_TIME_TO_LIVE_TIMEDELTA)
            self.assertEqual(msg.to,
                             self.MOCKED_TO)
            self.assertEqual(msg.transaction_partition_key,
                             self.MOCKED_TRANSACTION_PARTITION_KEY)
            self.assertDictEqual(msg.user_properties, {
                '$AzureWebJobsParentId': self.MOCKED_AZURE_PARTNER_ID,
                'x-opt-enqueue-sequence-number': 0
            })

    def test_servicebus_message_out_converter_encode_str(self):

        data = "dummy_string"

        result = azf_sb.ServiceBusMessageOutConverter.encode(
            obj=data, expected_type=None)

        self.assertEqual(result.type, "string")
        self.assertEqual(result.python_value, data)

    def test_servicebus_message_out_converter_encode_bytes(self):

        data = b"dummy_bytes"

        result = azf_sb.ServiceBusMessageOutConverter.encode(
            obj=data, expected_type=None)

        self.assertEqual(result.type, "bytes")
        self.assertEqual(result.python_value, data)

    def _generate_single_servicebus_data(self) -> meta.Datum:
        return meta.Datum(value=json.dumps({
            'lucky_number': 23
        }), type='json')

    def _generate_multiple_service_bus_data(self) -> meta.Datum:
        return meta.Datum(value=json.dumps([
            {'lucky_number': 23},
            {'lucky_number': 34},
            {'lucky_number': 45}
        ]), type='json')

    def _generate_single_trigger_metadata(self) -> Dict[str, meta.Datum]:
        """Generate a single ServiceBus message following
        https://docs.microsoft.com/en-us/azure/service-bus-messaging/
        service-bus-messages-payloads
        """

        mocked_metadata: Dict[str, meta.Datum] = {
            'ContentType': meta.Datum(
                self.MOCKED_CONTENT_TYPE, 'string'
            ),
            'CorrelationId': meta.Datum(
                self.MOCKED_CORROLATION_ID, 'string'
            ),
            'DeadLetterErrorDescription': meta.Datum(
                self.MOCKED_DEADLETTER_ERROR_DESCRIPTION, 'string'
            ),
            'DeadLetterReason': meta.Datum(
                self.MOCKED_DEADLETTER_REASON, 'string'
            ),
            'DeadLetterSource': meta.Datum(
                self.MOCKED_DEADLETTER_SOURCE, 'string'
            ),
            'DeliveryCount': meta.Datum(
                self.MOCKED_DELIVERY_COUNT, 'int'
            ),
            'EnqueuedSequenceNumber': meta.Datum(
                self.MOCKED_ENQUEUED_SEQUENCE_NUMBER, 'int'
            ),
            'EnqueuedTimeUtc': meta.Datum(
                self.MOCKED_ENQUEUE_TIME_UTC.isoformat(), 'string'
            ),
            'ExpiresAtUtc': meta.Datum(
                self.MOCKED_EXPIRY_AT_UTC.isoformat(), 'string'
            ),
            # 'ForcePersistence' not exposed yet, requires gRPC boolean passing
            'Label': meta.Datum(
                self.MOCKED_LABEL, 'string'
            ),
            'LockedUntil': meta.Datum(
                self.MOCKED_LOCKED_UNTIL.isoformat(), 'string'
            ),
            'LockToken': meta.Datum(
                self.MOCKED_LOCK_TOKEN, 'string'
            ),
            'MessageId': meta.Datum(
                self.MOCKED_MESSAGE_ID, 'string'
            ),
            'PartitionKey': meta.Datum(
                self.MOCKED_PARTITION_KEY, 'string'
            ),
            'ReplyTo': meta.Datum(
                self.MOCKED_REPLY_TO, 'string'
            ),
            'ReplyToSessionId': meta.Datum(
                self.MOCKED_REPLY_TO_SESSION_ID, 'string'
            ),
            'ScheduledEnqueueTimeUtc': meta.Datum(
                self.MOCKED_SCHEDULED_ENQUEUE_TIME_UTC.isoformat(), 'string'
            ),
            'SequenceNumber': meta.Datum(
                self.MOCKED_SEQUENCE_NUMBER, 'int'
            ),
            'SessionId': meta.Datum(
                self.MOCKED_SESSION_ID, 'string'
            ),
            'State': meta.Datum(
                self.MOCKED_STATE, 'int'
            ),
            'Subject': meta.Datum(
                self.MOCKED_SUBJECT, 'string'
            ),
            'TimeToLive': meta.Datum(
                self.MOCKED_TIME_TO_LIVE, 'string'
            ),
            'To': meta.Datum(
                self.MOCKED_TO, 'string'
            ),
            'TransactionPartitionKey': meta.Datum(
                self.MOCKED_TRANSACTION_PARTITION_KEY, 'string'
            )
        }
        mocked_metadata['MessageReceiver'] = meta.Datum(type='json', value='''
        {
            "RegisteredPlugins": [],
            "ReceiveMode": 0,
            "PrefetchCount": 0,
            "LastPeekedSequenceNumber": 0,
            "Path": "testqueue",
            "OperationTimeout": "00:01:00",
            "ServiceBusConnection": {
                "Endpoint": "sb://python-worker-36-sbns.servicebus.win.net",
                "OperationTimeout": "00:01:00",
                "RetryPolicy": {
                    "MinimalBackoff": "00:00:00",
                    "MaximumBackoff": "00:00:30",
                    "DeltaBackoff": "00:00:03",
                    "MaxRetryCount": 5,
                    "IsServerBusy": false,
                    "ServerBusyExceptionMessage": null
                },
                "TransportType": 0,
                "TokenProvider": {}
            },
            "IsClosedOrClosing": false,
            "ClientId": "MessageReceiver1testqueue",
            "RetryPolicy": {
                "MinimalBackoff": "00:00:00",
                "MaximumBackoff": "00:00:30",
                "DeltaBackoff": "00:00:03",
                "MaxRetryCount": 5,
                "IsServerBusy": false,
                "ServerBusyExceptionMessage": null
            }
        }''')
        mocked_metadata['ApplicationProperties'] = (
            meta.Datum(type='json', value='''
        {
            "application": "value"
        }
        '''))
        mocked_metadata['UserProperties'] = meta.Datum(type='json', value='''
        {
            "$AzureWebJobsParentId": "6ceef68b-0794-45dd-bb2e-630748515552",
            "x-opt-enqueue-sequence-number": 0
        }''')
        mocked_metadata['sys'] = meta.Datum(type='json', value='''
        {
            "MethodName": "ServiceBusSMany",
            "UtcNow": "2020-06-18T05:39:12.2860411Z",
            "RandGuid": "bb38deae-cc75-49f2-89f5-96ec6eb857db"
        }
        ''')
        return mocked_metadata

    def _generate_multiple_trigger_metadata(self) -> Dict[str, meta.Datum]:
        """Generate a metadatum containing 3 service bus messages which can be
        distingushed by enqueued_sequence_number
        """
        sb_a = self._generate_single_trigger_metadata()
        sb_b = self._generate_single_trigger_metadata()
        sb_c = self._generate_single_trigger_metadata()

        sb_a['MessageId'] = meta.Datum(self.MOCKED_MESSAGE_ID_A, 'string')
        sb_b['MessageId'] = meta.Datum(self.MOCKED_MESSAGE_ID_B, 'string')
        sb_c['MessageId'] = meta.Datum(self.MOCKED_MESSAGE_ID_C, 'string')

        combine_from = lambda key, et: self._zip(key, et, sb_a, sb_b, sb_c)

        mocked_metadata = {
            'ApplicationPropertiesArray': combine_from(
                'ApplicationProperties', 'json'
            ),
            'ContentTypeArray': combine_from(
                'ContentType', 'collection_string'
            ),
            'CorrelationIdArray': combine_from(
                'CorrelationId', 'collection_string'
            ),
            'DeadLetterErrorDescriptionArray': combine_from(
                'DeadLetterErrorDescription', 'collection_string'
            ),
            'DeadLetterReasonArray': combine_from(
                'DeadLetterReason', 'collection_string'
            ),
            'DeadLetterSourceArray': combine_from(
                'DeadLetterSource', 'collection_string'
            ),
            'EnqueuedSequenceNumberArray': combine_from(
                'EnqueuedSequenceNumber', 'collection_sint64'
            ),
            'EnqueuedTimeUtcArray': combine_from(
                'EnqueuedTimeUtc', 'json'
            ),
            'ExpiresAtUtcArray': combine_from(
                'ExpiresAtUtc', 'json'
            ),
            'LabelArray': combine_from(
                'Label', 'collection_string'
            ),
            'LockedUntilArray': combine_from(
                'LockedUntil', 'json'
            ),
            'LockTokenArray': combine_from(
                'LockToken', 'collection_string'
            ),
            'MessageIdArray': combine_from(
                'MessageId', 'collection_string'
            ),
            'PartitionKeyArray': combine_from(
                'PartitionKey', 'collection_string'
            ),
            'ReplyToArray': combine_from(
                'ReplyTo', 'collection_string'
            ),
            'ReplyToSessionIdArray': combine_from(
                'ReplyToSessionId', 'collection_string'
            ),
            'ScheduledEnqueueTimeUtcArray': combine_from(
                'ScheduledEnqueueTimeUtc', 'collection_string'
            ),
            'SessionIdArray': combine_from(
                'SessionId', 'collection_string'
            ),
            'SequenceNumberArray': combine_from(
                'SequenceNumber', 'collection_sint64'
            ),
            'StateArray': combine_from(
                'State', 'collection_sint64'
            ),
            'SubjectArray': combine_from(
                'Subject', 'collection_string'
            ),
            'TimeToLiveArray': combine_from(
                'TimeToLive', 'collection_string'
            ),
            'ToArray': combine_from(
                'To', 'collection_string'
            ),
            'TransactionPartitionKeyArray': combine_from(
                'TransactionPartitionKey', 'collection_string'
            ),
            'UserPropertiesArray': combine_from(
                'UserProperties', 'json'
            )
        }

        return mocked_metadata

    def _zip(self, key: str, expected_type: str,
             *args: List[Dict[str, meta.Datum]]) -> meta.Datum:
        """Combining multiple metadata into one:
        string -> collection_string
        bytes -> collection_bytes
        int -> collection_sint64
        sint64 -> collection_sint64
        json -> json (with array in it)
        """
        convertible = {
            'collection_string': CollectionString,
            'collection_bytes': CollectionBytes,
            'collection_sint64': CollectionSint64
        }

        datum_type = args[0][key].type
        if expected_type in convertible.keys():
            return meta.Datum(
                value=convertible[expected_type]([d[key].value for d in args]),
                type=expected_type
            )
        elif expected_type == 'json':
            if datum_type == 'json':
                value = json.dumps([json.loads(d[key].value) for d in args])
            else:
                value = json.dumps([d[key].value for d in args])
            return meta.Datum(
                value=value,
                type='json'
            )
        else:
            raise NotImplementedError(f'Unknown convertion {key}: '
                                      f'{datum_type} -> {expected_type}')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import NamedTuple, List
from .extension_hook_meta import ExtensionHookMeta


class FuncExtensionHooks(NamedTuple):
    """The definition of which type of function hooks are supported in SDK.
    ExtensionMeta will lookup the FuncExtension life-cycle type from here.
    """
    # The default value ([] empty list) is not being set here intentionally
    # since it is impacted by a Python bug https://bugs.python.org/issue33077.
    post_function_load: List[ExtensionHookMeta]
    pre_invocation: List[ExtensionHookMeta]
    post_invocation: List[ExtensionHookMeta]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging
import sys
import typing
from http.cookies import SimpleCookie

from azure.functions import _abc as azf_abc
from azure.functions import _http as azf_http
from . import meta
from ._jsonutils import json
from werkzeug.datastructures import Headers


class HttpRequest(azf_http.HttpRequest):
    """An HTTP request object."""

    __body_bytes: typing.Optional[bytes]
    __body_str: typing.Optional[str]

    def __init__(self,
                 method: str,
                 url: str, *,
                 headers: typing.Mapping[str, str],
                 params: typing.Mapping[str, str],
                 route_params: typing.Mapping[str, str],
                 body_type: str,
                 body: typing.Union[str, bytes]) -> None:

        body_str: typing.Optional[str] = None
        body_bytes: typing.Optional[bytes] = None
        if isinstance(body, str):
            body_str = body
            body_bytes = body_str.encode('utf-8')
        elif isinstance(body, bytes):
            body_bytes = body
        else:
            raise TypeError(
                f'unexpected HTTP request body type: {type(body).__name__}')

        super().__init__(method=method, url=url, headers=headers,
                         params=params, route_params=route_params,
                         body=body_bytes)

        self.__body_type = body_type
        self.__body_str = body_str
        self.__body_bytes = body_bytes

    def get_body(self) -> bytes:
        if self.__body_bytes is None:
            assert self.__body_str is not None
            self.__body_bytes = self.__body_str.encode('utf-8')
        return self.__body_bytes

    def get_json(self) -> typing.Any:
        if self.__body_type in ('json', 'string'):
            assert self.__body_str is not None
            return json.loads(self.__body_str)
        elif self.__body_bytes is not None:
            try:
                return json.loads(self.__body_bytes.decode('utf-8'))
            except ValueError as e:
                raise ValueError(
                    'HTTP request does not contain valid JSON data') from e
        else:
            raise ValueError(
                'Request body cannot be empty in JSON deserialization')


class HttpResponseConverter(meta.OutConverter, binding='http'):

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (azf_abc.HttpResponse, str))

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        if isinstance(obj, str):
            return meta.Datum(type='string', value=obj)

        if isinstance(obj, azf_abc.HttpResponse):
            status = obj.status_code
            headers: Headers = obj.headers

            if 'content-type' not in headers:
                if obj.mimetype.startswith('text/'):
                    ct = f'{obj.mimetype}; charset={obj.charset}'
                else:
                    ct = f'{obj.mimetype}'
                headers['content-type'] = ct

            body = obj.get_body()
            if body is not None:
                datum_body = meta.Datum(type='bytes', value=body)
            else:
                datum_body = meta.Datum(type='bytes', value=b'')

            cookies = None

            if sys.version_info.major == 3 and sys.version_info.minor <= 7:
                # SimpleCookie api in http.cookies - Python Standard Library
                # is not supporting 'samesite' in cookie attribute in python
                # 3.7 or below and would cause cookie parsing error
                # https://docs.python.org/3/library/http.cookies.html
                # ?msclkid=d78849ddcd7311ecadd81f2f51d08b8e
                logging.warning(
                    "Setting multiple 'Set-Cookie' response headers is not "
                    "supported in Azure Python Function with python version "
                    "3.7, please upgrade to python 3.8 or above.")
            else:
                if "Set-Cookie" in headers:
                    cookies = [SimpleCookie(cookie) for cookie in
                               headers.get_all('Set-Cookie')]
                    headers.pop("Set-Cookie")

            return meta.Datum(
                type='http',
                value=dict(
                    status_code=meta.Datum(type='string', value=str(status)),
                    headers={
                        n: meta.Datum(type='string', value=h)
                        for n, h in headers.items()
                    },
                    cookies=cookies,
                    body=datum_body,
                )
            )

        raise NotImplementedError


class HttpRequestConverter(meta.InConverter,
                           binding='httpTrigger', trigger=True):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, azf_abc.HttpRequest)

    @classmethod
    def decode(cls, data: meta.Datum, *,
               trigger_metadata) -> typing.Any:
        if data.type != 'http':
            raise NotImplementedError

        val = data.value

        return HttpRequest(
            method=val['method'].value,
            url=val['url'].value,
            headers={n: v.value for n, v in val['headers'].items()},
            params={n: v.value for n, v in val['query'].items()},
            route_params={n: v.value for n, v in val['params'].items()},
            body_type=val['body'].type,
            body=val['body'].value,
        )



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import DAPR_BINDING, DAPR_INVOKE, \
    DAPR_PUBLISH, DAPR_SECRET, DAPR_SERVICE_INVOCATION_TRIGGER, \
    DAPR_BINDING_TRIGGER, DAPR_STATE, DAPR_TOPIC_TRIGGER
from azure.functions.decorators.core import InputBinding, Trigger, DataType, \
    OutputBinding


class DaprServiceInvocationTrigger(Trigger):

    @staticmethod
    def get_binding_name() -> str:
        return DAPR_SERVICE_INVOCATION_TRIGGER

    def __init__(self,
                 name: str,
                 method_name: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.method_name = method_name
        super().__init__(name=name, data_type=data_type)


class DaprBindingTrigger(Trigger):

    @staticmethod
    def get_binding_name() -> str:
        return DAPR_BINDING_TRIGGER

    def __init__(self,
                 name: str,
                 binding_name: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.binding_name = binding_name
        super().__init__(name=name, data_type=data_type)


class DaprTopicTrigger(Trigger):

    @staticmethod
    def get_binding_name() -> str:
        return DAPR_TOPIC_TRIGGER

    def __init__(self,
                 name: str,
                 pub_sub_name: str,
                 topic: str,
                 route: Optional[str] = None,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.pub_sub_name = pub_sub_name
        self.topic = topic
        self.route = route
        super().__init__(name=name, data_type=data_type)


class DaprStateInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return DAPR_STATE

    def __init__(self,
                 name: str,
                 state_store: str,
                 key: str,
                 dapr_address: Optional[str],
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.state_store = state_store
        self.key = key
        self.dapr_address = dapr_address
        super().__init__(name=name, data_type=data_type)


class DaprSecretInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return DAPR_SECRET

    def __init__(self,
                 name: str,
                 secret_store_name: str,
                 key: str,
                 metadata: str,
                 dapr_address: Optional[str],
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.secret_store_name = secret_store_name
        self.key = key
        self.metadata = metadata
        self.dapr_address = dapr_address
        super().__init__(name=name, data_type=data_type)


class DaprStateOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return DAPR_STATE

    def __init__(self,
                 name: str,
                 state_store: str,
                 key: str,
                 dapr_address: Optional[str],
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.state_store = state_store
        self.key = key
        self.dapr_address = dapr_address
        super().__init__(name=name, data_type=data_type)


class DaprInvokeOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return DAPR_INVOKE

    def __init__(self,
                 name: str,
                 app_id: str,
                 method_name: str,
                 http_verb: str,
                 dapr_address: Optional[str],
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.app_id = app_id
        self.method_name = method_name
        self.http_verb = http_verb
        self.dapr_address = dapr_address
        super().__init__(name=name, data_type=data_type)


class DaprPublishOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return DAPR_PUBLISH

    def __init__(self,
                 name: str,
                 pub_sub_name: str,
                 topic: str,
                 dapr_address: Optional[str],
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.pub_sub_name = pub_sub_name
        self.topic = topic
        self.dapr_address = dapr_address
        super().__init__(name=name, data_type=data_type)


class DaprBindingOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return DAPR_BINDING

    def __init__(self,
                 name: str,
                 binding_name: str,
                 operation: str,
                 dapr_address: Optional[str],
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.binding_name = binding_name
        self.operation = operation
        self.dapr_address = dapr_address
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import collections.abc
import datetime
from typing import List, Dict, Any, Union, Optional

from azure.functions import _abc as azf_abc
from azure.functions import _queue as azf_queue

from . import meta
from ._jsonutils import json


class QueueMessage(azf_queue.QueueMessage):
    """An HTTP response object."""

    def __init__(self, *,
                 id=None, body=None,
                 dequeue_count=None,
                 expiration_time=None,
                 insertion_time=None,
                 time_next_visible=None,
                 pop_receipt=None):
        super().__init__(id=id, body=body, pop_receipt=pop_receipt)
        self.__dequeue_count = dequeue_count
        self.__expiration_time = expiration_time
        self.__insertion_time = insertion_time
        self.__time_next_visible = time_next_visible

    @property
    def dequeue_count(self):
        return self.__dequeue_count

    @property
    def expiration_time(self):
        return self.__expiration_time

    @property
    def insertion_time(self):
        return self.__insertion_time

    @property
    def time_next_visible(self):
        return self.__time_next_visible

    def __repr__(self) -> str:
        return (
            f'<azure.QueueMessage id={self.id} '
            f'dequeue_count={self.dequeue_count} '
            f'insertion_time={self.insertion_time} '
            f'expiration_time={self.expiration_time} '
            f'at 0x{id(self):0x}>'
        )


class QueueMessageInConverter(meta.InConverter,
                              binding='queueTrigger', trigger=True):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, azf_abc.QueueMessage)

    @classmethod
    def decode(cls, data: meta.Datum, *,
               trigger_metadata) -> Any:
        data_type = data.type

        if data_type in ['string', 'bytes']:
            body = data.value

        else:
            raise NotImplementedError(
                f'unsupported queue payload type: {data_type}')

        if trigger_metadata is None:
            raise NotImplementedError(
                'missing trigger metadata for queue input')

        return QueueMessage(
            id=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Id', python_type=str),
            body=body,
            dequeue_count=cls._decode_trigger_metadata_field(
                trigger_metadata, 'DequeueCount', python_type=int),
            expiration_time=cls._parse_datetime_metadata(
                trigger_metadata, 'ExpirationTime'),
            insertion_time=cls._parse_datetime_metadata(
                trigger_metadata, 'InsertionTime'),
            time_next_visible=cls._parse_datetime_metadata(
                trigger_metadata, 'NextVisibleTime'),
            pop_receipt=cls._decode_trigger_metadata_field(
                trigger_metadata, 'PopReceipt', python_type=str)
        )


class QueueMessageOutConverter(meta.OutConverter, binding='queue'):

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        valid_types = (azf_abc.QueueMessage, str, bytes)
        return (
            meta.is_iterable_type_annotation(pytype, valid_types)
            or (isinstance(pytype, type) and issubclass(pytype, valid_types))
        )

    @classmethod
    def encode(cls, obj: Any, *,
               expected_type: Optional[type]) -> meta.Datum:
        if isinstance(obj, str):
            return meta.Datum(type='string', value=obj)

        elif isinstance(obj, bytes):
            return meta.Datum(type='bytes', value=obj)

        elif isinstance(obj, azf_queue.QueueMessage):
            return meta.Datum(
                type='json',
                value=json.dumps({
                    'id': obj.id,
                    'body': obj.get_body().decode('utf-8'),
                })
            )

        elif isinstance(obj, collections.abc.Iterable):
            msgs: List[Union[str, Dict]] = []
            for item in obj:
                if isinstance(item, str):
                    msgs.append(item)
                elif isinstance(item, azf_queue.QueueMessage):
                    msgs.append({
                        'id': item.id,
                        'body': item.get_body().decode('utf-8')
                    })
                else:
                    raise NotImplementedError(
                        'invalid data type in output '
                        'queue message list: {}'.format(type(item)))

            return meta.Datum(
                type='json',
                value=json.dumps(msgs)
            )

        raise NotImplementedError

    @classmethod
    def _format_datetime(cls, dt: Optional[datetime.datetime]):
        if dt is None:
            return None
        else:
            return dt.isoformat()



import json
from typing import List

from azure.functions.decorators.utils import StringifyEnumJsonEncoder


class CollectionBytes:
    """The CollectionBytes class is used for generating a mock
    'collection_bytes' meta.Datum in testing. The common usage of it is
    new_datum = meta.Datum(type='collection_bytes',
                           value=CollectionBytes([b'1', b'2']))
    """
    def __init__(self, data: List[bytes]):
        self.bytes = data


class CollectionString:
    """The CollectionString class is used for generating a mock
    'collection_string' meta.Datum in testing. The common usage of it is
    new_datum = meta.Datum(type='collection_string',
                           value=CollectionString(['a', 'b']))
    """
    def __init__(self, data: List[str]):
        self.string = data


class CollectionSint64:
    """The CollectionSint64 class is used for generating a mock
    'collection_sint64' meta.Datum in testing. The common usage of it is
    new_datum = meta.Datum(type='collection_sint64',
                           value=CollectionSint64([1, 2]))
    """
    def __init__(self, data: List[int]):
        self.sint64 = data


def assert_json(self, func, expected_dict):
    self.assertEqual(json.dumps(json.loads(str(func)), sort_keys=True,
                                cls=StringifyEnumJsonEncoder),
                     json.dumps(expected_dict, sort_keys=True,
                                cls=StringifyEnumJsonEncoder))



import typing

from . import meta


class MCPToolTriggerConverter(meta.InConverter, binding='mcpToolTrigger',
                              trigger=True):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (str, dict, bytes))

    @classmethod
    def has_implicit_output(cls) -> bool:
        return True

    @classmethod
    def decode(cls, data: meta.Datum, *, trigger_metadata):
        """
        Decode incoming MCP tool request data.
        Returns the raw data in its native format (string, dict, bytes).
        """
        # Handle different data types appropriately
        if data.type == 'json':
            # If it's already parsed JSON, use the value directly
            return data.value
        elif data.type == 'string':
            # If it's a string, use it as-is
            return data.value
        elif data.type == 'bytes':
            return data.value
        else:
            # Fallback to python_value for other types
            return data.python_value if hasattr(data, 'python_value') else data.value

    @classmethod
    def encode(cls, obj: typing.Any, *, expected_type: typing.Optional[type] = None):
        """
        Encode the return value from MCP tool functions.
        MCP tools typically return string responses.
        """
        if obj is None:
            return meta.Datum(type='string', value='')
        elif isinstance(obj, str):
            return meta.Datum(type='string', value=obj)
        elif isinstance(obj, (bytes, bytearray)):
            return meta.Datum(type='bytes', value=bytes(obj))
        else:
            # Convert other types to string
            return meta.Datum(type='string', value=str(obj))



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.core import Trigger, \
    InputBinding, OutputBinding, DataType


class GenericInputBinding(InputBinding):

    @staticmethod
    def get_binding_name():
        pass

    def __init__(self,
                 name: str,
                 type: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        super().__init__(name=name, data_type=data_type, type=type)


class GenericOutputBinding(OutputBinding):

    @staticmethod
    def get_binding_name():
        pass

    def __init__(self,
                 name: str,
                 type: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        super().__init__(name=name, data_type=data_type, type=type)


class GenericTrigger(Trigger):

    @staticmethod
    def get_binding_name():
        pass

    def __init__(self,
                 name: str,
                 type: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        super().__init__(name=name, data_type=data_type, type=type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from azure.functions.extension.app_extension_hooks import AppExtensionHooks
import os
import unittest
from unittest.mock import MagicMock, patch
from logging import Logger

from azure.functions.extension import FunctionExtensionException
from azure.functions.extension.app_extension_base import AppExtensionBase
from azure.functions.extension.func_extension_base import FuncExtensionBase
from azure.functions.extension.extension_meta import ExtensionMeta
from azure.functions.extension.extension_scope import ExtensionScope
from azure.functions.extension.extension_hook_meta import ExtensionHookMeta
from azure.functions.extension.func_extension_hooks import FuncExtensionHooks
from azure.functions._abc import Context


class TestExtensionMeta(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self._instance = ExtensionMeta

    def tearDown(self) -> None:
        super().tearDown()
        self._instance._app_exts = None
        self._instance._func_exts.clear()
        self._instance._info.clear()

    def test_app_extension_should_register_to_app_exts(self):
        """When defining an application extension, it should be registered
        to application extension set in ExtensionMeta
        """
        # Define a new AppExtension
        NewAppExtension = _generate_new_app_extension(self._instance)

        # Check if the extension is actuall being loaded
        registered_post_function_load_exts = (
            self._instance._app_exts.post_function_load_app_level
        )
        self.assertEqual(
            registered_post_function_load_exts[0].ext_impl,
            NewAppExtension.post_function_load_app_level
        )

    def test_func_extension_should_register_to_func_exts(self):
        """When instantiating a function extension, it should be registered
        to function extension set in ExtensionMeta
        """
        # Define a new function extension
        NewFuncExtension = _generate_new_func_extension_class(self._instance)

        # Follow line should be executed from HttpTrigger/__init__.py script
        # Instantiate a new function extension
        ext_instance = NewFuncExtension()
        registered_post_function_load_exts = (
            self._instance._func_exts['httptrigger'].post_function_load[0]
        )
        self.assertEqual(
            registered_post_function_load_exts.ext_impl,
            ext_instance.post_function_load
        )

    def test_app_extension_base_should_not_be_registered(self):
        """When defining the app extension base, it should not be registerd
        since the base is not an actual application extension.
        """
        class AppExtensionBase(metaclass=self._instance):
            _scope = ExtensionScope.APPLICATION

        self.assertIsNone(self._instance._app_exts)

    def test_func_extension_base_should_not_be_registered(self):
        """When instantiating the func extension base, it should not be
        registered since the base is not an actual function extension.
        """
        class FuncExtensionBase(metaclass=self._instance):
            _scope = ExtensionScope.FUNCTION

        # Follow line should be executed from HttpTrigger/__init__.py script
        # Instantiate a new function extension base
        FuncExtensionBase()
        self.assertEqual(len(self._instance._func_exts), 0)

    def test_app_extension_instantiation_should_throw_error(self):
        """Application extension is operating on a class level, shouldn't be
        instantiate by trigger script
        """
        # Define a new AppExtension
        NewAppExtension = _generate_new_app_extension(self._instance)

        # Try instantiate the extension but it should fail
        with self.assertRaises(FunctionExtensionException):
            NewAppExtension()

    def test_invalid_scope_extension_instantiation_should_throw_error(self):
        """If the _scope is not defined in an extension, it is most likely
        an invalid extension
        """
        class InvalidExtension(metaclass=self._instance):
            pass

        with self.assertRaises(FunctionExtensionException):
            InvalidExtension()

    def test_get_function_hooks(self):
        """If a specific extension is registered in to a function, it should
        be able to retrieve it
        """
        extension_hook = self._instance._func_exts.setdefault(
            'httptrigger', FuncExtensionHooks(
                post_function_load=[],
                pre_invocation=[],
                post_invocation=[]
            )
        )

        extension_hook.post_function_load.append(
            ExtensionHookMeta(ext_impl=lambda: 'hello', ext_name='world')
        )

        hook = self._instance.get_function_hooks('HttpTrigger')
        self.assertEqual(hook.post_function_load[0].ext_impl(), 'hello')
        self.assertEqual(hook.post_function_load[0].ext_name, 'world')

    def test_get_application_hooks(self):
        """The application extension should be stored in self._app_hooks
        """
        hook_obj = AppExtensionHooks(
            post_function_load_app_level=[],
            pre_invocation_app_level=[],
            post_invocation_app_level=[]
        )
        self._instance._app_exts = hook_obj
        hooks = self._instance.get_application_hooks()
        self.assertEqual(id(hook_obj), id(hooks))

    def test_get_registered_extensions_json_empty(self):
        """Ensure the get extension json will return empty if there's nothing
        registered"""
        info_json = self._instance.get_registered_extensions_json()
        self.assertEqual(info_json, r'{}')

    def test_get_registered_extensions_json_function_ext(self):
        """Ensure the get extension json will return function ext info"""
        # Define a new function extension
        NewFuncExtension = _generate_new_func_extension_class(self._instance)

        # Instantiate the function extension as in a customer's code
        NewFuncExtension()
        info_json = self._instance.get_registered_extensions_json()
        self.assertEqual(
            info_json,
            r'{"FuncExtension": {"HttpTrigger": ["NewFuncExtension"]}}'
        )

    def test_get_registered_extension_json_application_ext(self):
        """Ensure the get extension json will return application ext info"""
        # Register a new application extension
        _generate_new_app_extension(self._instance)

        # The registration should be tracked in the info
        info_json = self._instance.get_registered_extensions_json()
        self.assertEqual(
            info_json,
            r'{"AppExtension": ["NewAppExtension"]}'
        )

    def test_get_extension_scope(self):
        """Test if ExtensionScope is properly retrieved"""
        # Register a new application extension
        NewAppExtension = _generate_new_app_extension(self._instance)

        scope = self._instance._get_extension_scope(NewAppExtension)
        self.assertEqual(scope, ExtensionScope.APPLICATION)

    def test_get_extenison_scope_not_set(self):
        """Test if ExtensionScope should be unknown when empty"""
        class InvalidExtension(metaclass=self._instance):
            pass

        scope = self._instance._get_extension_scope(InvalidExtension)
        self.assertEqual(scope, ExtensionScope.UNKNOWN)

    def test_set_hooks_for_function(self):
        """Instantiating a function extension will register the life-cycle
        hooks
        """
        # Define a new function extension
        NewFuncExtension = _generate_new_func_extension_class(self._instance)

        # Instantiate this as in HttpTrigger/__init__.py customer's code
        ext_instance = NewFuncExtension()
        self._instance._set_hooks_for_function('HttpTrigger', ext_instance)
        meta = self._instance._func_exts['httptrigger'].post_function_load[0]

        # Check extension name
        self.assertEqual(meta.ext_name, 'NewFuncExtension')

        # Check if the extension is executable
        meta.ext_impl(
            function_name='HttpTrigger',
            function_directory='/home/site/wwwroot/HttpTrigger'
        )
        self.assertTrue(ext_instance._post_function_load_executed)

    def test_set_hooks_for_application(self):
        """Create an application extension class will register the life-cycle
        hooks
        """
        # Register a new application extension
        NewAppExtension = _generate_new_app_extension(self._instance)

        self._instance._set_hooks_for_application(NewAppExtension)
        meta = self._instance._app_exts.post_function_load_app_level[0]

        # Check extension name
        self.assertEqual(meta.ext_name, 'NewAppExtension')

        # Check if extension is initialized and executable
        meta.ext_impl(function_name="HttpTrigger",
                      function_directory="/home/site/wwwroot")
        self.assertTrue(NewAppExtension._post_function_load_app_level_executed)

    def test_register_function_extension(self):
        """After intiializing, function extension should be recorded in
        func_exts and _info
        """
        # Define a new function extension
        NewFuncExtension = _generate_new_func_extension_class(self._instance)

        # The following line should be called by customer
        ext_instance = NewFuncExtension()
        self._instance._register_function_extension(ext_instance)

        # Check _func_exts should have lowercased trigger name
        self.assertIn('httptrigger', self._instance._func_exts)

        # Check _info should record the function extension
        self.assertIn(
            'HttpTrigger',
            self._instance._info.get('FuncExtension', {})
        )

    def test_register_application_extension(self):
        """After creating an application extension class, it should be recorded
        in app_exts and _info
        """
        # Register a new application extension
        _generate_new_app_extension(self._instance)

        # Check _app_exts should trigger_hook
        self.assertEqual(
            self._instance._app_exts.post_function_load_app_level[0].ext_name,
            'NewAppExtension'
        )

        # Check _info should record the application extension
        self.assertIn(
            'NewAppExtension', self._instance._info.get('AppExtension', [])
        )


class TestFuncExtensionBase(unittest.TestCase):

    def setUp(self):
        self.mock_script_root = os.path.join('/', 'home', 'site', 'wwwroot')
        self.mock_file_path = os.path.join(
            self.mock_script_root, 'HttpTrigger', '__init__.py'
        )
        extension_os_environ = os.environ.copy()
        extension_os_environ['AzureWebJobsScriptRoot'] = self.mock_script_root
        self.patch_os_environ = patch.dict('os.environ', extension_os_environ)
        self.patch_os_environ.start()

    def tearDown(self) -> None:
        self.patch_os_environ.stop()
        ExtensionMeta._info.clear()
        ExtensionMeta._func_exts.clear()
        ExtensionMeta._app_exts = None

    def test_new_extension_not_implement_init_should_fail(self):
        """An extension without __init__ should not be initialized
        """
        class NewExtension(FuncExtensionBase):
            pass

        # Initialize new extension without file path
        with self.assertRaises(TypeError):
            NewExtension()

    def test_new_extension_not_passing_filename_should_fail(self):
        """Instantiate an extension without __file__ in parameter should fail
        """
        class NewExtension(FuncExtensionBase):
            def __init__(self, file_path: str):
                # Initialize without super().__init__(file_path) fails
                super().__init__()

        # Initialize new extension without file path passing to super
        with self.assertRaises(TypeError):
            NewExtension('not_passed_to_super')

    def test_new_extension_invalid_path_should_fail(self):
        """The trigger name is derived from file script name. If an invalid
        path is provided, should throw exception.
        """
        class NewExtension(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

        # Customer try to register extension with invalid path name.
        # This should be pointing to a script __init__.py instead of a folder.
        with self.assertRaises(FunctionExtensionException):
            NewExtension('some_invalid_path')

    def test_new_extension_should_be_invalid_in_root_folder(self):
        """Function trigger path /home/site/wwwroot/<trigger_name>/__init__.py,
        If the path does not match this pattern, should throw error
        """
        class NewExtension(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

        # Customer try to register extension with /home/site/wwwroot.
        # This should be pointing to a script __init__.py instead of a folder.
        with self.assertRaises(FunctionExtensionException):
            NewExtension(self.mock_script_root)

    def test_new_extension_should_be_invalid_in_other_folder(self):
        """Function trigger path /home/site/wwwroot/<trigger_name>/__init__.py,
        If the path is not in /home/site/wwwroot, should throw error
        """
        class NewExtension(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

        # Customer try to register extension with /some/other/path.
        # This should be pointing to a script __init__.py instead of a folder.
        with self.assertRaises(FunctionExtensionException):
            NewExtension(os.path.join('/', 'some', 'other', 'path'))

    def test_new_extension_initialize_with_correct_path(self):
        """Instantiate an extension with __file__ in parameter should succeed
        """
        class NewExtension(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

        ext_instance = NewExtension(self.mock_file_path)
        self.assertEqual(ext_instance._trigger_name, 'HttpTrigger')

    def test_new_extension_should_succeed_with_submodule(self):
        """Instantiate an extension with __file__ in parameter should succeed
        from a trigger subfolder
        """
        class NewExtension(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

        # Customer try to register extension with /home/site/wwwroot.
        # This should be pointing to a script __init__.py instead of a folder.
        ext_instance = NewExtension(
            os.path.join(
                self.mock_script_root, 'HttpTrigger', 'SubModule',
                '__init__.py'
            )
        )
        self.assertEqual(ext_instance._trigger_name, 'HttpTrigger')

    def test_extension_registration(self):
        """Instantiate an extension with full life-cycle hooks support
        should be registered into _func_exts
        """
        class NewExtensionBeforeInvocation(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

            def post_function_load(self,
                                   function_name: str,
                                   function_directory: str,
                                   *args,
                                   **kwargs) -> None:
                print('ok_post_function_load')

            def pre_invocation(self, logger: Logger, context: Context,
                               *args, **kwargs) -> None:
                logger.info('ok_pre_invocation')

            def post_invocation(self, logger: Logger, context: Context,
                                *args, **kwargs) -> None:
                logger.info('ok_post_invocation')

        # Instantiate Extension
        ext_instance = NewExtensionBeforeInvocation(self.mock_file_path)

        # Check function hooks registration
        hooks = ExtensionMeta.get_function_hooks('HttpTrigger')
        self.assertIsInstance(hooks, FuncExtensionHooks)

        # Check post_function_load
        hook_meta = hooks.post_function_load[0]
        self.assertEqual(hook_meta.ext_name, ext_instance.__class__.__name__)
        self.assertEqual(hook_meta.ext_impl, ext_instance.post_function_load)

        # Check pre_invocation_hook
        hook_meta = hooks.pre_invocation[0]
        self.assertEqual(hook_meta.ext_name, ext_instance.__class__.__name__)
        self.assertEqual(hook_meta.ext_impl, ext_instance.pre_invocation)

        # Check post_invocation_hook
        hook_meta = hooks.post_invocation[0]
        self.assertEqual(hook_meta.ext_name, ext_instance.__class__.__name__)
        self.assertEqual(hook_meta.ext_impl, ext_instance.post_invocation)

    def test_partial_registration(self):
        """Instantiate an extension with full life-cycle hooks support
        should be registered into _func_exts
        """
        # Define extension with partial hooks support (e.g. post_invocation)
        class NewExtensionBeforeInvocation(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)

            def post_invocation(self, logger: Logger, context: Context,
                                *args, **kwargs) -> None:
                logger.info('ok_post_invocation')

        # Instantiate Extension
        ext_instance = NewExtensionBeforeInvocation(self.mock_file_path)

        # Check post_invocation hook registration
        hooks = ExtensionMeta.get_function_hooks('HttpTrigger')
        hook_meta = hooks.post_invocation[0]
        self.assertIsInstance(hooks, FuncExtensionHooks)
        self.assertEqual(hook_meta.ext_name, ext_instance.__class__.__name__)
        self.assertEqual(hook_meta.ext_impl, ext_instance.post_invocation)

    def test_extension_method_should_be_executed(self):
        """Ensure the life-cycle hook execution should happen
        """
        # Define extension with partial hooks support (e.g. post_invocation)
        class NewExtensionBeforeInvocation(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)
                self.is_post_invocation_executed = False

            def post_invocation(self, logger: Logger, context: Context,
                                *args, **kwargs) -> None:
                logger.info('ok_post_invocation')
                self.is_post_invocation_executed = True

        # Instantiate Extension
        ext_instance = NewExtensionBeforeInvocation(self.mock_file_path)

        # Check post_invocation hook invocation
        mock_logger = MagicMock()
        hooks = ExtensionMeta.get_function_hooks('HttpTrigger')
        self.assertFalse(ext_instance.is_post_invocation_executed)
        hooks.post_invocation[0].ext_impl(mock_logger, {})
        self.assertTrue(ext_instance.is_post_invocation_executed)

    def test_registration_should_lowercase_the_trigger_name(self):
        """The ExtensionMeta should not be case sensitive
        """
        class NewExtensionBeforeInvocation(FuncExtensionBase):
            def __init__(self, file_name: str):
                super().__init__(file_name)

            def pre_invocation(self, logger: Logger, context: Context,
                               *args, **kwargs) -> None:
                logger.info('ok_pre_invocation')

        NewExtensionBeforeInvocation(self.mock_file_path)

        # Check if the hooks can be retrieved from lower-cased trigger name
        self.assertIsNotNone(ExtensionMeta._func_exts.get('httptrigger'))

    def test_extension_invocation_should_have_logger(self):
        """Test if the extension can use the logger
        """
        class NewExtensionBeforeAndAfter(FuncExtensionBase):
            def __init__(self, trigger_name: str):
                super().__init__(trigger_name)

            def pre_invocation(self, logger: Logger, context: Context,
                               *args, **kwargs) -> None:
                logger.info('ok_pre_invocation')

        # Register extension in customer's
        NewExtensionBeforeAndAfter(self.mock_file_path)
        hooks = ExtensionMeta.get_function_hooks('HttpTrigger')

        # Check if the hook implementation executes
        mock_logger = MagicMock()
        hooks.pre_invocation[0].ext_impl(logger=mock_logger, context={})
        mock_logger.info.assert_called_with('ok_pre_invocation')

    def test_two_extensions_on_same_trigger(self):
        """Test if two extensions can be registered on the same trigger
        """
        class NewFuncExtension1(FuncExtensionBase):
            def __init__(self, trigger_name: str):
                super().__init__(trigger_name)

            def pre_invocation(self, logger: Logger, context: Context,
                               function_name: str, function_directory: str,
                               *args, **kwargs):
                return 'ok_before1'

        class NewFuncExtension2(FuncExtensionBase):
            def __init__(self, trigger_name: str):
                super().__init__(trigger_name)

            def pre_invocation(self, logger: Logger, context: Context,
                               function_name: str, function_directory: str,
                               *args, **kwargs):
                return 'ok_before2'

        # Check if both extensions are registered under the same hook
        NewFuncExtension1(self.mock_file_path)
        NewFuncExtension2(self.mock_file_path)
        hooks = ExtensionMeta.get_function_hooks('HttpTrigger')

        # Check if the before invocation hook matches metadata
        extension_names = map(
            lambda x: getattr(x, 'ext_name'),
            hooks.pre_invocation
        )
        self.assertIn('NewFuncExtension1', extension_names)
        self.assertIn('NewFuncExtension2', extension_names)

        # Check if the extension can be executed correctly
        extension_implementations = list(map(
            lambda x: getattr(x, 'ext_impl'),
            hooks.pre_invocation
        ))
        self.assertEqual(
            extension_implementations[0](
                logger=None, context={}, function_name='HttpTrigger',
                function_directory='/home/site/wwwroot/HttpTrigger'
            ), 'ok_before1'
        )
        self.assertEqual(
            extension_implementations[1](
                logger=None, context={}, function_name='HttpTrigger',
                function_directory='/home/site/wwwroot/HttpTrigger'
            ), 'ok_before2'
        )

    def test_backward_compatilbility_less_arguments(self):
        """Test if the existing extension implemented the interface with
        less arguments
        """
        class ExtensionWithLessArgument(FuncExtensionBase):
            def __init__(self, file_path: str):
                super().__init__(file_path)
                self.executed = False

            # Drop arguments
            def pre_invocation(self):
                self.executed = True

        # Check if the before invocation hook matches metadata
        ext_instance = ExtensionWithLessArgument(self.mock_file_path)
        hooks = ExtensionMeta.get_function_hooks('HttpTrigger')

        # Check if implementation works
        hook_meta = hooks.pre_invocation[0]
        self.assertEqual(hook_meta.ext_name, 'ExtensionWithLessArgument')

        # Check if the hook implementation executes
        hook_meta.ext_impl()
        self.assertTrue(ext_instance.executed)


class TestAppExtensionBase(unittest.TestCase):

    def setUp(self):
        super().setUp()
        self._instance = ExtensionMeta
        self.patch_os_environ = patch.dict('os.environ', os.environ.copy())
        self.patch_os_environ.start()

    def tearDown(self) -> None:
        super().tearDown()
        self.patch_os_environ.stop()
        ExtensionMeta._app_exts = None
        ExtensionMeta._info.clear()
        ExtensionMeta._func_exts.clear()

    def test_empty_app_extension_should_pass(self):
        """An application extension can be registered directly since it never
        gets instantiate. Defining a new AppExtension should not raise an
        exception.
        """
        class NewEmptyAppExtension(AppExtensionBase):
            pass

    def test_init_method_should_be_called(self):
        """An application extension's init() classmethod should be called
        when the class is created"""
        # Define new an application extension
        NewAppExtension = _generate_new_app_extension(self._instance)

        # Ensure the init() method is executed
        self.assertTrue(NewAppExtension._init_executed)

    def test_extension_registration(self):
        """The life-cycles implementations in extension should be automatically
        registered in class creation
        """
        # Define new an application extension
        NewAppExtension = _generate_new_app_extension(self._instance)

        # Check app hooks registration
        hooks = self._instance.get_application_hooks()
        self.assertIsInstance(hooks, AppExtensionHooks)

        # Check post_function_load_app_level
        hook_meta = hooks.post_function_load_app_level[0]
        self.assertEqual(hook_meta.ext_name, 'NewAppExtension')
        self.assertEqual(hook_meta.ext_impl,
                         NewAppExtension.post_function_load_app_level)

        # Execute post_function_load_app_level
        hook_meta.ext_impl(
            logger=None, context={}, function_name='HttpTrigger',
            function_directory='/home/site/wwwroot/HttpTrigger'
        )
        self.assertTrue(
            NewAppExtension._post_function_load_app_level_executed
        )

        # Check pre_invocation_app_level
        hook_meta = hooks.pre_invocation_app_level[0]
        self.assertEqual(hook_meta.ext_name, 'NewAppExtension')
        self.assertEqual(hook_meta.ext_impl,
                         NewAppExtension.pre_invocation_app_level)

        # Execute pre_invocation_app_level
        hook_meta.ext_impl(logger=None, context={})
        self.assertTrue(
            NewAppExtension._pre_invocation_app_level
        )

        # Check post_invocation_app_level
        hook_meta = hooks.post_invocation_app_level[0]
        self.assertEqual(hook_meta.ext_name, 'NewAppExtension')
        self.assertEqual(hook_meta.ext_impl,
                         NewAppExtension.post_invocation_app_level)

        # Exectue post_invocation_app_level
        hook_meta.ext_impl(logger=None, context={})
        self.assertTrue(
            NewAppExtension._post_invocation_app_level
        )


def _generate_new_func_extension_class(
    metaclass: type,
    trigger_name: str = 'HttpTrigger'
):
    class NewFuncExtension(metaclass=metaclass):
        _scope = ExtensionScope.FUNCTION

        def __init__(self):
            self._trigger_name = trigger_name
            self._post_function_load_executed = False
            self._pre_invocation_executed = False
            self._post_invocation_executed = False

        def post_function_load(self, function_name, function_directory):
            self._post_function_load_executed = True

        def pre_invocation(self, logger, context):
            self._pre_invocation_executed = True

        def post_invocation(self, logger, context):
            self._post_invocation_executed = True

    return NewFuncExtension


def _generate_new_app_extension(metaclass: type):
    class NewAppExtension(metaclass=metaclass):
        _scope = ExtensionScope.APPLICATION

        _init_executed = False
        _post_function_load_app_level_executed = False
        _pre_invocation_app_level = False
        _post_invocation_app_level = False

        @classmethod
        def init(cls):
            cls._init_executed = True

        @classmethod
        def post_function_load_app_level(cls,
                                         function_name,
                                         function_directory,
                                         *args,
                                         **kwargs):
            cls._post_function_load_app_level_executed = True

        @classmethod
        def pre_invocation_app_level(cls, logger, context,
                                     *args, **kwargs):
            cls._pre_invocation_app_level = True

        @classmethod
        def post_invocation_app_level(cls, logger, context,
                                      *args, **kwargs):
            cls._post_invocation_app_level = True

    return NewAppExtension



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from abc import ABC
import inspect
import json
import unittest
from unittest import mock
from unittest.mock import patch

from azure.functions import WsgiMiddleware, AsgiMiddleware
from azure.functions.decorators.constants import HTTP_OUTPUT, HTTP_TRIGGER, \
    TIMER_TRIGGER
from azure.functions.decorators.core import DataType, AuthLevel, \
    BindingDirection, SCRIPT_FILE_NAME
from azure.functions.decorators.function_app import (
    BindingApi, FunctionBuilder, FunctionApp, Function, Blueprint,
    DecoratorApi, AsgiFunctionApp, SettingsApi, WsgiFunctionApp,
    HttpFunctionsAuthLevelMixin, FunctionRegister, TriggerApi,
    ExternalHttpFunctionApp
)
from azure.functions.decorators.http import HttpTrigger, HttpOutput, \
    HttpMethod
from azure.functions.decorators.retry_policy import RetryPolicy
from test_core import DummyTrigger
from tests.utils.testutils import assert_json


class TestFunction(unittest.TestCase):

    def setUp(self):
        def dummy():
            return "dummy"

        self.dummy = dummy
        self.func = Function(self.dummy, "dummy.py")

    def test_function_creation(self):
        self.assertEqual(self.func.get_user_function(), self.dummy)
        self.assertEqual(self.func.function_script_file, "dummy.py")

    def test_add_trigger(self):
        with self.assertRaises(ValueError) as err:
            trigger1 = HttpTrigger(name="req1", methods=(HttpMethod.GET,),
                                   data_type=DataType.UNDEFINED,
                                   auth_level=AuthLevel.ANONYMOUS,
                                   route="dummy")

            trigger2 = HttpTrigger(name="req2", methods=(HttpMethod.GET,),
                                   data_type=DataType.UNDEFINED,
                                   auth_level=AuthLevel.ANONYMOUS,
                                   route="dummy")

            self.func.add_trigger(trigger1)
            self.assertEqual(len(self.func.get_bindings()), 1)
            self.assertEqual(self.func.get_bindings()[0], trigger1)
            self.func.add_trigger(trigger2)
            self.assertEqual(err.exception,
                             "A trigger was already registered to this "
                             "function. Adding another trigger is not the "
                             "correct behavior as a function can only have one"
                             " trigger. Existing registered trigger "
                             f"is {trigger1} and New trigger "
                             f"being added is {trigger2}")

    def test_function_creation_with_binding_and_trigger(self):
        output = HttpOutput(name="out", data_type=DataType.UNDEFINED)
        trigger = HttpTrigger(name="req",
                              methods=(HttpMethod.GET, HttpMethod.POST),
                              data_type=DataType.UNDEFINED,
                              auth_level=AuthLevel.ANONYMOUS, route="dummy")
        self.func.add_binding(output)
        self.func.add_trigger(trigger)

        self.assertEqual(self.func.get_function_name(), "dummy")
        self.assertEqual(self.func.get_user_function(), self.dummy)
        assert_json(self, self.func, {"scriptFile": "dummy.py",
                                      "bindings": [
                                          {
                                              "type": HTTP_OUTPUT,
                                              "direction":
                                                  BindingDirection.OUT,
                                              "name": "out",
                                              "dataType": DataType.UNDEFINED
                                          },
                                          {
                                              "authLevel": AuthLevel.ANONYMOUS,
                                              "type": HTTP_TRIGGER,
                                              "direction": BindingDirection.IN,
                                              "name": "req",
                                              "dataType": DataType.UNDEFINED,
                                              "route": "dummy",
                                              "methods": [HttpMethod.GET,
                                                          HttpMethod.POST]
                                          }
                                      ]
                                      })

        raw_bindings = self.func.get_raw_bindings()
        self.assertEqual(len(raw_bindings), 2)
        raw_output_binding = raw_bindings[0]
        raw_trigger = raw_bindings[1]

        self.assertEqual(json.loads(raw_output_binding),
                         json.loads(
                             '{"direction": "OUT", "dataType": "UNDEFINED", '
                             '"type": "http", "name": "out"}'))

        self.assertEqual(json.loads(raw_trigger),
                         json.loads(
                             '{"direction": "IN", "dataType": "UNDEFINED", '
                             '"type": "httpTrigger", "authLevel": '
                             '"ANONYMOUS", "route": "dummy", "methods": ['
                             '"GET", "POST"], "name": "req"}'))


class TestFunctionBuilder(unittest.TestCase):

    def test_function_builder_creation(self):
        def test_function_builder_creation():
            return "dummy"

        self.dummy = test_function_builder_creation
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        self.assertTrue(callable(self.fb))
        func = getattr(self.fb, "_function")
        self.assertEqual(self.fb._function.function_script_file, "dummy.py")
        self.assertEqual(func.get_user_function(), self.dummy)

    def test_validate_function_missing_trigger(self):
        def test_validate_function_missing_trigger():
            return "dummy"

        self.dummy = test_validate_function_missing_trigger
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        with self.assertRaises(ValueError) as err:
            #  self.fb.configure_function_name('dummy').build()
            self.fb.build()

        self.assertEqual(err.exception.args[0],
                         "Function test_validate_function_missing_trigger"
                         " does not have a trigger. A valid function must have"
                         " one and only one trigger registered.")

    def test_validate_function_trigger_not_in_bindings(self):
        def test_validate_function_trigger_not_in_bindings():
            return "dummy"

        self.dummy = test_validate_function_trigger_not_in_bindings
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        trigger = HttpTrigger(name='req', methods=(HttpMethod.GET,),
                              data_type=DataType.UNDEFINED,
                              auth_level=AuthLevel.ANONYMOUS,
                              route='dummy')
        with self.assertRaises(ValueError) as err:
            self.fb.add_trigger(trigger)
            getattr(self.fb, "_function").get_bindings().clear()
            self.fb.build()

        self.assertEqual(
            err.exception.args[0],
            f"Function test_validate_function_trigger_not_in_bindings"
            f" trigger {trigger} not present"
            f" in bindings {[]}")

    def test_validate_function_working(self):
        def test_validate_function_working():
            return "dummy"

        self.dummy = test_validate_function_working
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        trigger = HttpTrigger(name='req', methods=(HttpMethod.GET,),
                              data_type=DataType.UNDEFINED,
                              auth_level=AuthLevel.ANONYMOUS,
                              route='test_validate_function_working')
        self.fb.add_trigger(trigger)
        self.fb.build()

    def test_build_function_http_route_default(self):
        def test_build_function_http_route_default():
            return "dummy"

        self.dummy = test_build_function_http_route_default
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        trigger = HttpTrigger(name='req', methods=(HttpMethod.GET,),
                              data_type=DataType.UNDEFINED,
                              auth_level=AuthLevel.ANONYMOUS)
        self.fb.add_trigger(trigger)
        func = self.fb.build()

        self.assertEqual(func.get_trigger().route,
                         "test_build_function_http_route_default")

    def test_build_function_with_bindings(self):
        def test_build_function_with_bindings():
            return "dummy"

        self.dummy = test_build_function_with_bindings
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        test_trigger = HttpTrigger(name='req', methods=(HttpMethod.GET,),
                                   data_type=DataType.UNDEFINED,
                                   auth_level=AuthLevel.ANONYMOUS)
        test_input = HttpOutput(name='out', data_type=DataType.UNDEFINED)

        func = self.fb.add_trigger(
            test_trigger).add_binding(test_input).build()

        self.assertEqual(func.get_function_name(),
                         "test_build_function_with_bindings")
        assert_json(self, func, {
            "scriptFile": "dummy.py",
            "bindings": [
                {
                    "authLevel": AuthLevel.ANONYMOUS,
                    "type": HTTP_TRIGGER,
                    "direction": BindingDirection.IN,
                    "name": "req",
                    "dataType": DataType.UNDEFINED,
                    "route": "test_build_function_with_bindings",
                    "methods": [
                        HttpMethod.GET
                    ]
                },
                {
                    "type": HTTP_OUTPUT,
                    "direction": BindingDirection.OUT,
                    "name": "out",
                    "dataType": DataType.UNDEFINED
                }
            ]
        })

    def test_build_function_with_function_app_auth_level(self):
        def test_build_function_with_function_app_auth_level():
            return "dummy"

        self.dummy = test_build_function_with_function_app_auth_level
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        trigger = HttpTrigger(name='req', methods=(HttpMethod.GET,),
                              data_type=DataType.UNDEFINED)
        self.fb.add_trigger(trigger)
        func = self.fb.build(auth_level=AuthLevel.ANONYMOUS)

        self.assertEqual(func.get_trigger().auth_level, AuthLevel.ANONYMOUS)

    def test_build_function_with_retry_policy_setting(self):
        def test_build_function_with_retry_policy_setting():
            return "dummy"

        self.dummy = test_build_function_with_retry_policy_setting
        self.fb = FunctionBuilder(self.dummy, "dummy.py")

        setting = RetryPolicy(strategy="exponential", max_retry_count="2",
                              minimum_interval="1", maximum_interval="5")
        trigger = HttpTrigger(
            name='req', methods=(HttpMethod.GET,),
            data_type=DataType.UNDEFINED,
            auth_level=AuthLevel.ANONYMOUS,
            route='test_build_function_with_retry_policy_setting')
        self.fb.add_trigger(trigger)
        self.fb.add_setting(setting)
        func = self.fb.build()

        self.assertEqual(func.get_settings_dict("retry_policy"),
                         {'setting_name': 'retry_policy',
                          'strategy': 'exponential', 'max_retry_count': '2',
                          'minimum_interval': '1', 'maximum_interval': '5'})

    def test_unique_method_names(self):
        app = FunctionApp()

        @app.schedule(arg_name="name", schedule="10****")
        def test_unique_method_names(name: str):
            return name

        @app.schedule(arg_name="name", schedule="10****")
        def test_unique_method_names2(name: str):
            return name

        functions = app.get_functions()
        self.assertEqual(len(functions), 2)

        self.assertEqual(functions[0].get_function_name(),
                         "test_unique_method_names")
        self.assertEqual(functions[1].get_function_name(),
                         "test_unique_method_names2")

    def test_unique_function_names(self):
        app = FunctionApp()

        @app.function_name("test_unique_function_names")
        @app.schedule(arg_name="name", schedule="10****")
        def test_unique_function_names(name: str):
            return name

        @app.function_name("test_unique_function_names2")
        @app.schedule(arg_name="name", schedule="10****")
        def test_unique_function_names2(name: str):
            return name

        functions = app.get_functions()
        self.assertEqual(len(functions), 2)

        self.assertEqual(functions[0].get_function_name(),
                         "test_unique_function_names")
        self.assertEqual(functions[1].get_function_name(),
                         "test_unique_function_names2")

    def test_same_method_names(self):
        app = FunctionApp()

        @app.schedule(arg_name="name", schedule="10****") # NoQA
        def test_same_method_names(name: str): # NoQA
            return name

        @app.schedule(arg_name="name", schedule="10****") # NoQA
        def test_same_method_names(name: str): # NoQA
            return name

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(err.exception.args[0],
                         "Function test_same_method_names does not have"
                         " a unique function name."
                         " Please change @app.function_name()"
                         " or the function method name to be unique.")

    def test_same_function_names(self):
        app = FunctionApp()

        @app.function_name("test_same_function_names") # NoQA
        @app.schedule(arg_name="name", schedule="10****")
        def test_same_function_names(name: str):
            return name

        @app.function_name("test_same_function_names") # NoQA
        @app.schedule(arg_name="name", schedule="10****")
        def test_same_function_names(name: str): # NoQA
            return name

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(err.exception.args[0],
                         "Function test_same_function_names does not have"
                         " a unique function name."
                         " Please change @app.function_name()"
                         " or the function method name to be unique.")

    def test_same_function_name_different_method_name(self):
        app = FunctionApp()

        @app.function_name("test_same_function_name_different_method_name")
        @app.schedule(arg_name="name", schedule="10****")
        def test_same_function_name_different_method_name(name: str):
            return name

        @app.function_name("test_same_function_name_different_method_name")
        @app.schedule(arg_name="name", schedule="10****")
        def test_same_function_name_different_method_name2(name: str):
            return name

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(
            err.exception.args[0],
            "Function test_same_function_name_different_method_name"
            " does not have a unique function name."
            " Please change @app.function_name()"
            " or the function method name to be unique.")

    def test_same_function_and_method_name(self):
        app = FunctionApp()

        @app.function_name("test_same_function_and_method_name")
        @app.schedule(arg_name="name", schedule="10****")
        def test_same_function_and_method_name2(name: str):
            return name

        @app.schedule(arg_name="name", schedule="10****")
        def test_same_function_and_method_name(name: str):
            return name

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(err.exception.args[0],
                         "Function test_same_function_and_method_name"
                         " does not have a unique function name."
                         " Please change @app.function_name()"
                         " or the function method name to be unique.")

    def test_blueprint_unique_method_names(self):
        app = FunctionApp()

        @app.schedule(arg_name="name", schedule="10****")
        def test_blueprint_unique_method_names(name: str):
            return name

        bp = Blueprint()

        @bp.schedule(arg_name="name", schedule="10****")
        def test_blueprint_unique_method_names2(name: str):
            return name

        app.register_blueprint(bp)

        functions = app.get_functions()
        self.assertEqual(len(functions), 2)

        self.assertEqual(functions[0].get_function_name(),
                         "test_blueprint_unique_method_names")
        self.assertEqual(functions[1].get_function_name(),
                         "test_blueprint_unique_method_names2")

    def test_blueprint_unique_function_names(self):
        app = FunctionApp()

        @app.function_name("test_blueprint_unique_function_names")
        @app.schedule(arg_name="name", schedule="10****")
        def test_blueprint_unique_function_names(name: str):
            return name

        bp = Blueprint()

        @bp.function_name("test_blueprint_unique_function_names2")
        @bp.schedule(arg_name="name", schedule="10****")
        def test_blueprint_unique_function_names2(name: str):
            return name

        app.register_blueprint(bp)

        functions = app.get_functions()
        self.assertEqual(len(functions), 2)

        self.assertEqual(functions[0].get_function_name(),
                         "test_blueprint_unique_function_names")
        self.assertEqual(functions[1].get_function_name(),
                         "test_blueprint_unique_function_names2")

    def test_blueprint_same_method_names(self):
        app = FunctionApp()

        @app.schedule(arg_name="name", schedule="10****") # NoQA
        def test_blueprint_same_method_names(name: str): # NoQA
            return name

        bp = Blueprint()

        @bp.schedule(arg_name="name", schedule="10****") # NoQA
        def test_blueprint_same_method_names(name: str): # NoQA
            return name

        app.register_blueprint(bp)

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(err.exception.args[0],
                         "Function test_blueprint_same_method_names"
                         " does not have a unique function name."
                         " Please change @app.function_name()"
                         " or the function method name to be unique.")

    def test_blueprint_same_function_names(self):
        app = FunctionApp()

        @app.function_name("test_blueprint_same_function_names") # NoQA
        @app.schedule(arg_name="name", schedule="10****")
        def test_blueprint_same_function_names(name: str):  # NoQA
            return name

        bp = Blueprint()

        @bp.function_name("test_blueprint_same_function_names") # NoQA
        @bp.schedule(arg_name="name", schedule="10****")
        def test_blueprint_same_function_names(name: str):  # NoQA
            return name

        app.register_blueprint(bp)

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(err.exception.args[0],
                         "Function test_blueprint_same_function_names"
                         " does not have a unique function name. Please change"
                         " @app.function_name() or the function method name"
                         " to be unique.")

    def test_blueprint_same_function_name_different_method_name(self):
        app = FunctionApp()

        @app.function_name(
            "test_blueprint_same_function_name_different_method_name")
        @app.schedule(arg_name="name", schedule="10****")
        def test_blueprint_same_function_name_different_method_name(name: str):
            return name

        bp = Blueprint()

        @bp.function_name(
            "test_blueprint_same_function_name_different_method_name")
        @bp.schedule(arg_name="name", schedule="10****")
        def test_blueprint_same_function_name_different_method_name2(
                name: str):
            return name

        app.register_blueprint(bp)

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(
            err.exception.args[0],
            "Function test_blueprint_same_function_name_different_method_name"
            " does not have a unique function name. Please change"
            " @app.function_name() or the function method name to be unique.")

    def test_blueprint_same_function_and_method_name(self):
        app = FunctionApp()

        @app.function_name("test_blueprint_same_function_and_method_name")
        @app.schedule(arg_name="name", schedule="10****")
        def test_blueprint_same_function_and_method_name2(name: str):
            return name

        bp = Blueprint()

        @bp.schedule(arg_name="name", schedule="10****")
        def test_blueprint_same_function_and_method_name(name: str):
            return name

        app.register_blueprint(bp)

        with self.assertRaises(ValueError) as err:
            app.get_functions()
        self.assertEqual(
            err.exception.args[0],
            "Function test_blueprint_same_function_and_method_name"
            " does not have a unique function name."
            " Please change @app.function_name() or the function"
            " method name to be unique.")

    def test_user_function_is_directly_callable_no_args(self):
        def test_validate_function_working_no_args():
            return "dummy"

        self.dummy = test_validate_function_working_no_args
        self.fb = FunctionBuilder(self.dummy, "dummy.py")
        self.assertEqual(self.fb(), "dummy")

    def test_user_function_is_directly_callable_args(self):
        def test_validate_function_working_sum_args(arg1: int, arg2: int):
            return arg1 + arg2

        self.dummy = test_validate_function_working_sum_args
        self.fb = FunctionBuilder(self.dummy, "dummy.py")
        self.assertEqual(self.fb(1, 2), 3)


class TestScaffold(unittest.TestCase):
    def setUp(self):
        class DummyApp(DecoratorApi):
            def dummy_trigger(self, name: str):
                @self._configure_function_builder
                def wrap(fb):
                    def decorator():
                        fb.add_trigger(trigger=DummyTrigger(name=name))
                        return fb

                    return decorator()

                return wrap

        self.dummy = DummyApp()

    def test_has_app_script_file(self):
        self.assertTrue(self.dummy.app_script_file, SCRIPT_FILE_NAME)

    def test_has_function_builders(self):
        self.assertEqual(self.dummy._function_builders, [])

    def test_dummy_app_trigger(self):
        @self.dummy.dummy_trigger(name="dummy")
        def test_dummy_app_trigger():
            return "dummy"

        self.assertEqual(len(self.dummy._function_builders), 1)
        func = self.dummy._function_builders[0].build()
        self.assertEqual(func.get_function_name(), "test_dummy_app_trigger")
        self.assertEqual(func.get_function_json(),
                         '{"scriptFile": "function_app.py", "bindings": [{'
                         '"direction": "IN", "dataType": "UNDEFINED", '
                         '"type": "Dummy", "name": "dummy"}]}')


class TestFunctionApp(unittest.TestCase):
    def setUp(self):
        def dummy_func():
            pass

        self.dummy_func = dummy_func
        self.func_app = FunctionApp()

    def test_default_auth_level_function(self):
        self.assertEqual(self.func_app.auth_level, AuthLevel.FUNCTION)

    def test_default_script_file_path(self):
        self.assertEqual(self.func_app.app_script_file, SCRIPT_FILE_NAME)

    def test_auth_level(self):
        self.func_app = FunctionApp(http_auth_level='ANONYMOUS')
        self.assertEqual(self.func_app.auth_level, AuthLevel.ANONYMOUS)

    def test_get_no_functions(self):
        self.assertEqual(self.func_app.app_script_file, "function_app.py")
        self.assertEqual(self.func_app.get_functions(), [])

    def test_invalid_decorated_type(self):
        with self.assertRaises(ValueError) as err:
            self.func_app._validate_type(object())

        self.assertEqual(err.exception.args[0], "Unsupported type for "
                                                "function app decorator "
                                                "found.")

    def test_callable_decorated_type(self):
        fb = self.func_app._validate_type(self.dummy_func)
        self.assertTrue(isinstance(fb, FunctionBuilder))
        self.assertEqual(fb._function.get_user_function(), self.dummy_func)

    def test_function_builder_decorated_type(self):
        fb = FunctionBuilder(self.dummy_func, "dummy.py")
        self.func_app._function_builders.append(fb)

        fb = self.func_app._validate_type(fb)
        self.assertTrue(isinstance(fb, FunctionBuilder))
        self.assertEqual(fb._function.get_user_function(), self.dummy_func)

    @mock.patch('azure.functions.decorators.function_app.AsgiFunctionApp'
                '._add_http_app')
    def test_add_asgi(self, add_http_app_mock):
        mock_asgi_app = object()
        AsgiFunctionApp(app=mock_asgi_app, function_name='test_add_asgi')

        add_http_app_mock.assert_called_once()

        self.assertIsInstance(add_http_app_mock.call_args[0][0],
                              AsgiMiddleware)

    @mock.patch('azure.functions.decorators.function_app.WsgiFunctionApp'
                '._add_http_app')
    def test_add_wsgi(self, add_http_app_mock):
        mock_wsgi_app = object()
        WsgiFunctionApp(app=mock_wsgi_app, function_name='test_add_wsgi')

        add_http_app_mock.assert_called_once()
        self.assertIsInstance(add_http_app_mock.call_args[0][0],
                              WsgiMiddleware)

    def test_extends_required_classes(self):
        self.assertTrue(issubclass(ExternalHttpFunctionApp, FunctionRegister))
        self.assertTrue(issubclass(ExternalHttpFunctionApp, TriggerApi))
        self.assertTrue(issubclass(ExternalHttpFunctionApp, SettingsApi))
        self.assertTrue(issubclass(ExternalHttpFunctionApp, BindingApi))
        self.assertTrue(issubclass(ExternalHttpFunctionApp, ABC))
        self.assertTrue(issubclass(AsgiFunctionApp, ExternalHttpFunctionApp))
        self.assertTrue(issubclass(WsgiFunctionApp, ExternalHttpFunctionApp))

    def test_add_asgi_app(self):
        self._test_http_external_app(AsgiFunctionApp(
            app=object(),
            function_name='test_add_asgi_app'),
            True,
            function_name='test_add_asgi_app')

    def test_add_wsgi_app(self):
        self._test_http_external_app(WsgiFunctionApp(
            app=object(),
            function_name='test_add_wsgi_app'),
            False,
            function_name='test_add_wsgi_app')

    def test_register_function_app_error(self):
        with self.assertRaises(TypeError) as err:
            FunctionApp().register_functions(FunctionApp())

        self.assertEqual(err.exception.args[0],
                         "functions can not be type of FunctionRegister!")

    def test_register_blueprint(self):
        bp = Blueprint()

        @bp.schedule(arg_name="name", schedule="10****")
        def test_register_blueprint(name: str):
            return "hello"

        app = FunctionApp()
        app.register_blueprint(bp)

        self.assertEqual(len(app.get_functions()), 1)
        self.assertEqual(app.auth_level, AuthLevel.FUNCTION)
        self.assertEqual(app.app_script_file, SCRIPT_FILE_NAME)

    def test_register_app_auth_level(self):
        bp = Blueprint()

        @bp.route("name")
        def test_register_app_auth_level(name: str):
            return "hello"

        app = FunctionApp(http_auth_level=AuthLevel.ANONYMOUS)
        app.register_blueprint(bp)

        functions = app.get_functions()
        self.assertEqual(len(functions), 1)
        self.assertEqual(functions[0].get_trigger().auth_level,
                         AuthLevel.ANONYMOUS)

    def test_default_function_http_type(self):
        app = FunctionApp(http_auth_level=AuthLevel.ANONYMOUS)

        @app.route("name")
        def hello(name: str):
            return "hello"

        funcs = app.get_functions()
        self.assertEqual(len(funcs), 1)

        func = funcs[0]
        self.assertEqual(func.http_type, 'function')

    def test_set_http_type(self):
        app = FunctionApp(http_auth_level=AuthLevel.ANONYMOUS)

        @app.route("name1")
        @app.http_type("dummy1")
        def test_set_http_type(name: str):
            return "hello"

        @app.route("name2")
        @app.http_type("dummy2")
        def test_set_http_type2(name: str):
            return "hello"

        funcs = app.get_functions()
        self.assertEqual(len(funcs), 2)

        func1 = funcs[0]
        self.assertEqual(func1.http_type, 'dummy1')
        func2 = funcs[1]
        self.assertEqual(func2.http_type, 'dummy2')

    def test_decorator_api_basic_props(self):
        class DummyFunctionApp(DecoratorApi):
            pass

        app = DummyFunctionApp()

        self.assertEqual(app.app_script_file, SCRIPT_FILE_NAME)
        self.assertIsNotNone(getattr(app, "_validate_type", None))
        self.assertIsNotNone(getattr(app, "_configure_function_builder", None))

    def test_http_functions_auth_level_mixin(self):
        class DummyFunctionApp(HttpFunctionsAuthLevelMixin):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)

        self.assertTrue(hasattr(app, "auth_level"))
        self.assertEqual(app.auth_level, AuthLevel.ANONYMOUS)

    def test_function_register_basic_props(self):
        class DummyFunctionApp(FunctionRegister):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)

        self.assertEqual(app.app_script_file, SCRIPT_FILE_NAME)
        self.assertIsNotNone(getattr(app, "_validate_type", None))
        self.assertIsNotNone(getattr(app, "_configure_function_builder", None))
        self.assertIsNone(getattr(app, "_require_auth_level"))
        self.assertTrue(hasattr(app, "auth_level"))
        self.assertEqual(app.auth_level, AuthLevel.ANONYMOUS)

    def test_function_register_http_function_app(self):
        class DummyFunctionApp(FunctionRegister, TriggerApi):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)

        @app.route("name1")
        def hello1(name: str):
            return "hello"

        @app.schedule(arg_name="name", schedule="10****")
        def hello2(name: str):
            return "hello"

        @app.route("name1")
        def hello3(name: str):
            return "hello"

        self.assertIsNone(app._require_auth_level, None)
        app.get_functions()
        self.assertTrue(app._require_auth_level)

    def test_function_register_non_http_function_app(self):
        class DummyFunctionApp(FunctionRegister, TriggerApi):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)
        blueprint = Blueprint()

        @blueprint.schedule(arg_name="name", schedule="10****")
        def test_function_register_non_http_function_app(name: str):
            return name

        app.register_blueprint(blueprint)

        app.get_functions()
        self.assertFalse(app._require_auth_level)

    def test_blueprints_with_function_name(self):
        class DummyFunctionApp(FunctionRegister, TriggerApi):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)
        blueprint = Blueprint()

        @blueprint.function_name("timer_function")
        @blueprint.schedule(arg_name="name", schedule="10****")
        def hello(name: str):
            return name

        app.register_blueprint(blueprint)

        functions = app.get_functions()
        self.assertEqual(len(functions), 1)

        setting = functions[0].get_setting("function_name")

        self.assertEqual(setting.get_settings_value("function_name"),
                         "timer_function")

    def test_legacy_blueprints_with_function_name(self):
        class LegacyBluePrint(TriggerApi, BindingApi):
            pass

        class DummyFunctionApp(FunctionRegister, TriggerApi):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)
        blueprint = LegacyBluePrint()

        @blueprint.function_name("test_legacy_blueprints_with_function_name")
        @blueprint.schedule(arg_name="name", schedule="10****")
        def hello(name: str):
            return name

        app.register_blueprint(blueprint)

        functions = app.get_functions()
        self.assertEqual(len(functions), 1)

        setting = functions[0].get_setting("function_name")

        self.assertEqual(setting.get_settings_value("function_name"),
                         "test_legacy_blueprints_with_function_name")

    def test_function_register_register_function_register_error(self):
        class DummyFunctionApp(FunctionRegister):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)
        with self.assertRaises(TypeError) as err:
            app.register_functions(app)

        self.assertEqual(err.exception.args[0],
                         'functions can not be type of FunctionRegister!')

    def test_function_register_register_functions_from_blueprint(self):
        class DummyFunctionApp(FunctionRegister):
            pass

        app = DummyFunctionApp(auth_level=AuthLevel.ANONYMOUS)
        blueprint = Blueprint()

        @blueprint.schedule(arg_name="name", schedule="10****")
        def test_function_register_register_functions_from_blueprint(
                name: str):
            return name

        app.register_blueprint(blueprint)

        functions = app.get_functions()
        self.assertEqual(len(functions), 1)

        trigger = functions[0].get_trigger()

        self.assertEqual(trigger.type, TIMER_TRIGGER)
        self.assertEqual(trigger.schedule, "10****")
        self.assertEqual(trigger.name, "name")
        self.assertEqual(
            functions[0].get_function_name(),
            "test_function_register_register_functions_from_blueprint")
        self.assertEqual(functions[0].get_user_function()("timer"), "timer")

    def test_asgi_function_app_default(self):
        app = AsgiFunctionApp(app=object())
        self.assertEqual(app.auth_level, AuthLevel.FUNCTION)

    def test_asgi_function_app_custom(self):
        app = AsgiFunctionApp(app=object(),
                              http_auth_level=AuthLevel.ANONYMOUS)
        self.assertEqual(app.auth_level, AuthLevel.ANONYMOUS)

    def test_asgi_function_app_is_http_function(self):
        app = AsgiFunctionApp(app=object())
        funcs = app.get_functions()

        self.assertEqual(len(funcs), 1)
        self.assertTrue(funcs[0].is_http_function())

    def test_wsgi_function_app_default(self):
        app = WsgiFunctionApp(app=object())
        self.assertEqual(app.auth_level, AuthLevel.FUNCTION)

    def test_wsgi_function_app_custom(self):
        app = WsgiFunctionApp(app=object(),
                              http_auth_level=AuthLevel.ANONYMOUS)
        self.assertEqual(app.auth_level, AuthLevel.ANONYMOUS)

    def test_wsgi_function_app_is_http_function(self):
        app = WsgiFunctionApp(
            app=object(),
            function_name='test_wsgi_function_app_is_http_function')
        funcs = app.get_functions()

        self.assertEqual(len(funcs), 1)
        self.assertTrue(funcs[0].is_http_function())

    def test_asgi_function_app_add_wsgi_app(self):
        with self.assertRaises(TypeError) as err:
            app = AsgiFunctionApp(app=object(),
                                  http_auth_level=AuthLevel.ANONYMOUS)
            app._add_http_app(WsgiMiddleware(object()))

        self.assertEqual(err.exception.args[0],
                         "Please pass AsgiMiddleware instance as parameter.")

    def test_wsgi_function_app_add_asgi_app(self):
        with self.assertRaises(TypeError) as err:
            app = WsgiFunctionApp(app=object(),
                                  http_auth_level=AuthLevel.ANONYMOUS)
            app._add_http_app(AsgiMiddleware(object()))

        self.assertEqual(err.exception.args[0],
                         "Please pass WsgiMiddleware instance as parameter.")

    @patch("azure.functions.decorators.function_app.ExternalHttpFunctionApp"
           ".__abstractmethods__", set())
    def test_external_http_function_app(self):
        with self.assertRaises(NotImplementedError):
            app = ExternalHttpFunctionApp(auth_level=AuthLevel.ANONYMOUS)
            app._add_http_app(AsgiMiddleware(object()))

    def _test_http_external_app(self, app, is_async, function_name):
        funcs = app.get_functions()
        self.assertEqual(len(funcs), 1)
        func = funcs[0]
        self.assertEqual(func.get_function_name(), function_name)
        raw_bindings = func.get_raw_bindings()
        raw_trigger = raw_bindings[0]
        raw_output_binding = raw_bindings[0]
        self.assertEqual(inspect.iscoroutinefunction(func.get_user_function()),
                         is_async)
        self.assertEqual(json.loads(raw_trigger),
                         json.loads(
                             '{"direction": "IN", "type": "httpTrigger", '
                             '"authLevel": "FUNCTION", "route": "/{*route}", '
                             '"methods": ["GET", "POST", "DELETE", "HEAD", '
                             '"PATCH", "PUT", "OPTIONS"], "name": "req"}'))
        self.assertEqual(json.loads(raw_output_binding), json.loads(
            '{"direction": "IN", "type": "httpTrigger", "authLevel": '
            '"FUNCTION", "methods": ["GET", "POST", "DELETE", "HEAD", '
            '"PATCH", "PUT", "OPTIONS"], "name": "req", "route": "/{'
            '*route}"}'))
        self.assertEqual(func.get_bindings_dict(), {
            "bindings": [
                {
                    "authLevel": AuthLevel.FUNCTION,
                    "direction": BindingDirection.IN,
                    "methods": [HttpMethod.GET, HttpMethod.POST,
                                HttpMethod.DELETE,
                                HttpMethod.HEAD,
                                HttpMethod.PATCH,
                                HttpMethod.PUT, HttpMethod.OPTIONS],
                    "name": "req",
                    "route": "/{*route}",
                    "type": HTTP_TRIGGER
                },
                {
                    "direction": BindingDirection.OUT,
                    "name": "$return",
                    "type": HTTP_OUTPUT
                }
            ]})


class TestFunctionRegister(unittest.TestCase):
    def test_validate_empty_dict(self):
        def dummy():
            return "dummy"

        test_func = Function(dummy, "dummy.py")
        fr = FunctionRegister(auth_level="ANONYMOUS")
        fr.validate_function_names(functions=[test_func])

    def test_validate_unique_names(self):
        def dummy():
            return "dummy"

        def dummy2():
            return "dummy"

        test_func = Function(dummy, "dummy.py")
        test_func2 = Function(dummy2, "dummy.py")

        fr = FunctionRegister(auth_level="ANONYMOUS")
        fr.validate_function_names(
            functions=[test_func, test_func2])

    def test_validate_non_unique_names(self):
        def dummy():
            return "dummy"

        test_func = Function(dummy, "dummy.py")
        test_func2 = Function(dummy, "dummy.py")

        fr = FunctionRegister(auth_level="ANONYMOUS")
        with self.assertRaises(ValueError) as err:
            fr.validate_function_names(functions=[test_func, test_func2])
            self.assertEqual(err.exception.args[0],
                             "Function dummy does not have"
                             " a unique function name."
                             " Please change @app.function_name()"
                             " or the function method name to be unique.")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import abc
import collections.abc
import datetime
import re
from typing import Dict, Optional, Union, Tuple, Mapping, Any

from ._jsonutils import json
from ._thirdparty import typing_inspect
from ._utils import (
    try_parse_datetime_with_formats,
    try_parse_timedelta_with_formats
)


def is_iterable_type_annotation(annotation: object, pytype: object) -> bool:
    is_iterable_anno = (
        typing_inspect.is_generic_type(annotation)
        and issubclass(typing_inspect.get_origin(annotation),
                       collections.abc.Iterable)
    )

    if not is_iterable_anno:
        return False

    args = typing_inspect.get_args(annotation)
    if not args:
        return False

    if isinstance(pytype, tuple):
        return any(isinstance(t, type) and issubclass(t, arg)
                   for t in pytype for arg in args)
    else:
        return any(isinstance(pytype, type) and issubclass(pytype, arg)
                   for arg in args)


class Datum:
    def __init__(self, value: Any, type: Optional[str]):
        self.value: Any = value
        self.type: Optional[str] = type

    @property
    def python_value(self) -> Any:
        if self.value is None or self.type is None:
            return None
        elif self.type in ('bytes', 'string', 'int', 'double'):
            return self.value
        elif self.type == 'json':
            return json.loads(self.value)
        elif self.type == 'collection_string':
            return [v for v in self.value.string]
        elif self.type == 'collection_bytes':
            return [v for v in self.value.bytes]
        elif self.type == 'collection_double':
            return [v for v in self.value.double]
        elif self.type == 'collection_sint64':
            return [v for v in self.value.sint64]
        else:
            return self.value

    @property
    def python_type(self) -> type:
        return type(self.python_value)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.value == other.value and self.type == other.type

    def __hash__(self):
        return hash((type(self), (self.value, self.type)))

    def __repr__(self):
        val_repr = repr(self.value)
        if len(val_repr) > 10:
            val_repr = val_repr[:10] + '...'
        return '<Datum {} {}>'.format(self.type, val_repr)


class _ConverterMeta(abc.ABCMeta):

    _bindings: Dict[str, type] = {}

    def __new__(mcls, name, bases, dct, *,
                binding: Optional[str],
                trigger: Optional[str] = None):
        cls = super().__new__(mcls, name, bases, dct)
        cls._trigger = trigger  # type: ignore
        if binding is None:
            return cls

        if binding in mcls._bindings:
            raise RuntimeError(
                f'cannot register a converter for {binding!r} binding: '
                f'another converter for this binding has already been '
                f'registered')

        mcls._bindings[binding] = cls
        if trigger is not None:
            mcls._bindings[trigger] = cls

        return cls

    @classmethod
    def get(cls, binding_name):
        return cls._bindings.get(binding_name)

    def has_trigger_support(cls) -> bool:
        return cls._trigger is not None  # type: ignore


class _BaseConverter(metaclass=_ConverterMeta, binding=None):

    @classmethod
    def _decode_typed_data(
            cls, data: Datum, *,
            python_type: Union[type, Tuple[type, ...]],
            context: str = 'data') -> Any:
        if data is None:
            return None

        data_type = data.type
        if data_type == 'json':
            result = json.loads(data.value)

        elif data_type == 'string':
            result = data.value

        elif data_type == 'int':
            result = data.value

        elif data_type == 'double':
            result = data.value

        elif data_type == 'collection_bytes':
            result = data.value

        elif data_type == 'collection_string':
            result = data.value

        elif data_type == 'collection_sint64':
            result = data.value

        elif data_type is None:
            return None

        else:
            raise ValueError(
                f'unsupported type of {context}: {data_type}')

        if not isinstance(result, python_type):
            if isinstance(python_type, (tuple, list, dict)):
                raise ValueError(
                    f'unexpected value type in {context}: '
                    f'{type(result).__name__}, expected one of: '
                    f'{", ".join(t.__name__ for t in python_type)}')
            else:
                try:
                    # Try coercing into the requested type
                    result = python_type(result)
                except (TypeError, ValueError) as e:
                    raise ValueError(
                        f'cannot convert value of {context} into '
                        f'{python_type.__name__}: {e}') from None

        return result

    @classmethod
    def _decode_trigger_metadata_field(
            cls, trigger_metadata: Mapping[str, Datum],
            field: str, *,
            python_type: Union[type, Tuple[type, ...]]) \
            -> Any:
        data = trigger_metadata.get(field)
        if data is None:
            return None
        else:
            return cls._decode_typed_data(
                data, python_type=python_type,
                context=f'field {field!r} in trigger metadata')

    @classmethod
    def _parse_datetime_metadata(
            cls, trigger_metadata: Mapping[str, Datum],
            field: str) -> Optional[datetime.datetime]:

        datetime_str = cls._decode_trigger_metadata_field(
            trigger_metadata, field, python_type=str)

        if datetime_str is None:
            return None
        else:
            return cls._parse_datetime(datetime_str)

    @classmethod
    def _parse_timedelta_metadata(
            cls, trigger_metadata: Mapping[str, Datum],
            field: str) -> Optional[datetime.timedelta]:

        timedelta_str = cls._decode_trigger_metadata_field(
            trigger_metadata, field, python_type=str)

        if timedelta_str is None:
            return None
        else:
            return cls._parse_timedelta(timedelta_str)

    @classmethod
    def _parse_datetime(
            cls, datetime_str: Optional[str]) -> Optional[datetime.datetime]:

        if not datetime_str:
            return None

        too_fractional = re.match(
            r'(.*\.\d{6})(\d+)(Z|[\+|-]\d{1,2}:\d{1,2}){0,1}', datetime_str)

        if too_fractional:
            # The supplied value contains seven digits in the
            # fractional second part, whereas Python expects
            # a maxium of six, so strip it.
            # https://github.com/Azure/azure-functions-python-worker/issues/269
            datetime_str = too_fractional.group(1) + (
                too_fractional.group(3) or '')

        # Try parse time
        utc_time, utc_time_error = cls._parse_datetime_utc(datetime_str)
        if not utc_time_error and utc_time:
            return utc_time.replace(tzinfo=datetime.timezone.utc)

        local_time, local_time_error = cls._parse_datetime_local(datetime_str)
        if not local_time_error and local_time:
            return local_time.replace(tzinfo=None)

        # Report error
        if utc_time_error:
            raise utc_time_error
        elif local_time_error:
            raise local_time_error
        else:
            return None

    @classmethod
    def _parse_timedelta(
        cls,
        timedelta_str: Optional[str]
    ) -> Optional[datetime.timedelta]:

        if not timedelta_str:
            return None

        # Try parse timedelta
        timedelta, td_error = cls._parse_timedelta_internal(timedelta_str)
        if timedelta is not None:
            return timedelta

        # Report error
        if td_error:
            raise td_error
        else:
            return None

    @classmethod
    def _parse_datetime_utc(
        cls,
        datetime_str: str
    ) -> Tuple[Optional[datetime.datetime], Optional[Exception]]:

        # UTC ISO 8601 assumed
        # 2018-08-07T23:17:57.461050Z
        utc_formats = [
            '%Y-%m-%dT%H:%M:%S+00:00',
            '%Y-%m-%dT%H:%M:%S-00:00',
            '%Y-%m-%dT%H:%M:%S.%f+00:00',
            '%Y-%m-%dT%H:%M:%S.%f-00:00',
            '%Y-%m-%dT%H:%M:%SZ',
            '%Y-%m-%dT%H:%M:%S.%fZ',

            '%m/%d/%Y %H:%M:%SZ',
            '%m/%d/%Y %H:%M:%S.%fZ',
            '%m/%d/%Y %H:%M:%S+00:00',
            '%m/%d/%Y %H:%M:%S-00:00',
            '%m/%d/%Y %H:%M:%S.%f+00:00',
            '%m/%d/%Y %H:%M:%S.%f-00:00',
        ]

        dt, _, excpt = try_parse_datetime_with_formats(
            datetime_str, utc_formats)

        if excpt is not None:
            return None, excpt
        return dt, None

    @classmethod
    def _parse_datetime_local(
        cls, datetime_str: str
    ) -> Tuple[Optional[datetime.datetime], Optional[Exception]]:
        """Parse a string into a datetime object, accepts following formats
        1. Without fractional seconds (e.g. 2018-08-07T23:17:57)
        2. With fractional seconds (e.g. 2018-08-07T23:17:57.461050)

        Parameters
        ----------
        datetime_str: str
            The string represents a datetime

        Returns
        -------
        Tuple[Optional[datetime.datetime], Optional[Exception]]
            If the datetime_str is None, will return None immediately.
            If the datetime_str can be parsed correctly, it will return as the
            first element in the tuple.
            If the datetime_str cannot be parsed with all attempts, it will
            return None in the first element, the exception in the second
            element.
        """

        local_formats = [
            '%Y-%m-%dT%H:%M:%S.%f',
            '%Y-%m-%dT%H:%M:%S',

            '%m/%d/%YT%H:%M:%S.%f',
            '%m/%d/%YT%H:%M:%S'
        ]

        dt, _, excpt = try_parse_datetime_with_formats(
            datetime_str, local_formats)

        if excpt is not None:
            return None, excpt
        return dt, None

    @classmethod
    def _parse_timedelta_internal(
        cls, timedelta_str: str
    ) -> Tuple[Optional[datetime.timedelta], Optional[Exception]]:
        """Parse a string into a timedelta object, accepts following formats
        1. HH:MM:SS (e.g. 12:34:56)
        2. MM:SS (e.g. 34:56)
        3. Pure integer as seconds (e.g. 5819)

        Parameters
        ----------
        timedelta_str: str
            The string represents a datetime

        Returns
        -------
        Tuple[Optional[datetime.timedelta], Optional[Exception]]
            If the timedelta_str is None, will return None immediately.
            If the timedelta_str can be parsed correctly, it will return as the
            first element in the tuple.
            If the timedelta_str cannot be parsed with all attempts, it will
            return None in the first element, the exception in the second
            element.
        """

        timedelta_formats = [
            '%H:%M:%S',
            '%M:%S',
            '%S'
        ]

        td, _, excpt = try_parse_timedelta_with_formats(
            timedelta_str, timedelta_formats)

        if td is not None:
            return td, None
        return None, excpt


class InConverter(_BaseConverter, binding=None):

    @classmethod
    @abc.abstractmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        pass

    @classmethod
    @abc.abstractmethod
    def decode(cls, data: Datum, *, trigger_metadata) -> Any:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def has_implicit_output(cls) -> bool:
        return False


class OutConverter(_BaseConverter, binding=None):

    @classmethod
    @abc.abstractmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        pass

    @classmethod
    @abc.abstractmethod
    def encode(cls, obj: Any, *,
               expected_type: Optional[type]) -> Optional[Datum]:
        raise NotImplementedError


def get_binding_registry():
    return _ConverterMeta



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict, Any, List, Union, Optional, Mapping

from azure.functions import _eventhub
from ._jsonutils import json

from . import meta


class EventHubConverter(meta.InConverter, meta.OutConverter,
                        binding='eventHub'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        valid_types = (_eventhub.EventHubEvent)
        return (
            meta.is_iterable_type_annotation(pytype, valid_types)
            or (isinstance(pytype, type) and issubclass(pytype, valid_types))
        )

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        valid_types = (str, bytes)
        return (
            meta.is_iterable_type_annotation(pytype, str)
            or (isinstance(pytype, type) and issubclass(pytype, valid_types))
        )

    @classmethod
    def decode(
        cls, data: meta.Datum, *, trigger_metadata
    ) -> Union[_eventhub.EventHubEvent, List[_eventhub.EventHubEvent]]:
        data_type = data.type

        if data_type in ['string', 'bytes', 'json']:
            return cls.decode_single_event(data, trigger_metadata)

        elif data_type in ['collection_bytes', 'collection_string']:
            return cls.decode_multiple_events(data, trigger_metadata)

        else:
            raise NotImplementedError(
                f'unsupported event data payload type: {data_type}')

    @classmethod
    def decode_single_event(cls, data,
                            trigger_metadata) -> _eventhub.EventHubEvent:
        if data.type in ['string', 'json']:
            body = data.value.encode('utf-8')

        elif data.type == 'bytes':
            body = data.value

        return _eventhub.EventHubEvent(body=body)

    @classmethod
    def decode_multiple_events(
            cls, data, trigger_metadata
    ) -> List[_eventhub.EventHubEvent]:
        if data.type == 'collection_bytes':
            parsed_data = data.value.bytes

        elif data.type == 'collection_string':
            parsed_data = data.value.string

        events = []
        for parsed_datum in parsed_data:
            event = _eventhub.EventHubEvent(body=parsed_datum)
            events.append(event)

        return events

    @classmethod
    def encode(cls, obj: Any, *,
               expected_type: Optional[type]
               ) -> meta.Datum:
        data = meta.Datum(type=None, value=None)

        if isinstance(obj, str):
            data = meta.Datum(type='string', value=obj)

        elif isinstance(obj, bytes):
            data = meta.Datum(type='bytes', value=obj)

        elif isinstance(obj, int):
            data = meta.Datum(type='int', value=obj)

        elif isinstance(obj, list):
            data = meta.Datum(type='json', value=json.dumps(obj))

        return data


class EventHubTriggerConverter(EventHubConverter,
                               binding='eventHubTrigger', trigger=True):
    @classmethod
    def decode(
        cls, data: meta.Datum, *, trigger_metadata: Mapping[str, meta.Datum]
    ) -> Union[_eventhub.EventHubEvent, List[_eventhub.EventHubEvent]]:
        data_type = data.type

        if cls._is_cardinality_one(trigger_metadata):
            return cls.decode_single_event(data, trigger_metadata)

        elif cls._is_cardinality_many(trigger_metadata):
            return cls.decode_multiple_events(data, trigger_metadata)

        else:
            raise NotImplementedError(
                f'unsupported event data payload type: {data_type}')

    @classmethod
    def decode_single_event(
        cls, data, trigger_metadata: Mapping[str, meta.Datum]
    ) -> _eventhub.EventHubEvent:
        if data.type in ['string', 'json']:
            body = data.value.encode('utf-8')

        elif data.type == 'bytes':
            body = data.value

        return _eventhub.EventHubEvent(
            body=body,
            trigger_metadata=trigger_metadata,
            enqueued_time=cls._parse_datetime_metadata(
                trigger_metadata, 'EnqueuedTimeUtc'),
            partition_key=cls._decode_trigger_metadata_field(
                trigger_metadata, 'PartitionKey', python_type=str),
            sequence_number=cls._decode_trigger_metadata_field(
                trigger_metadata, 'SequenceNumber', python_type=int),
            offset=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Offset', python_type=str),
            iothub_metadata=cls._decode_iothub_metadata(trigger_metadata)
        )

    @classmethod
    def decode_multiple_events(
            cls, data, trigger_metadata: Mapping[str, meta.Datum]
    ) -> List[_eventhub.EventHubEvent]:
        if data.type == 'collection_bytes':
            parsed_data = data.value.bytes

        elif data.type == 'collection_string':
            parsed_data = data.value.string

        # Input Trigger IotHub Event
        elif data.type == 'json':
            parsed_data = json.loads(data.value)

        sys_props = trigger_metadata.get('SystemPropertiesArray')

        parsed_sys_props: List[Any] = []
        if sys_props is not None:
            parsed_sys_props = json.loads(sys_props.value)

        if len(parsed_data) != len(parsed_sys_props):
            raise AssertionError('Number of bodies and metadata mismatched')

        events = []
        for i in range(len(parsed_data)):
            enqueued_time = parsed_sys_props[i].get('EnqueuedTimeUtc')
            partition_key = cls.encode(
                parsed_sys_props[i].get('PartitionKey'),
                expected_type=str)
            sequence_number = cls.encode(
                parsed_sys_props[i].get('SequenceNumber'),
                expected_type=int)
            offset = cls.encode(
                parsed_sys_props[i].get('Offset'),
                expected_type=int)

            event = _eventhub.EventHubEvent(
                body=cls._marshall_event_body(parsed_data[i], data.type),
                trigger_metadata=trigger_metadata,
                enqueued_time=cls._parse_datetime(enqueued_time),
                partition_key=cls._decode_typed_data(
                    partition_key, python_type=str),
                sequence_number=cls._decode_typed_data(
                    sequence_number, python_type=int),
                offset=cls._decode_typed_data(
                    offset, python_type=int),
                iothub_metadata=cls._extract_iothub_from_dict(
                    parsed_sys_props[i])
            )

            events.append(event)

        return events

    @classmethod
    def _marshall_event_body(cls, parsed_data, data_type):
        # In IoTHub, when setting the eventhub using cardinality = 'many'
        # The data is wrapped inside a json (e.g. '[{ "device-id": "1" }]')

        # Previously, since the IoTHub events has a 'json' datatype,
        # it is handled as single_event by mistake and our users handle the
        # data parsing. And we want to keep the same behavior here.
        if data_type == 'json':
            return json.dumps(parsed_data).encode('utf-8')
        elif data_type == 'bytes':
            return parsed_data
        elif data_type == 'string':
            return parsed_data.encode('utf-8')
        elif data_type == 'collection_bytes':
            return parsed_data
        elif data_type == 'collection_string':
            return parsed_data.encode('utf-8')

        return parsed_data

    @classmethod
    def _decode_iothub_metadata(
            cls, trigger_metadata) -> Dict[str, str]:
        # Try extracting iothub_metadata from trigger_metadata
        iothub_metadata = cls._extract_iothub_from_trigger_metadata(
            trigger_metadata)

        # Try extracting iothub_metadata from SystemProperties
        if not iothub_metadata and trigger_metadata.get('SystemProperties'):
            iothub_metadata = cls._extract_iothub_from_system_properties(
                trigger_metadata['SystemProperties'].value)

        return iothub_metadata

    @classmethod
    def _extract_iothub_from_trigger_metadata(
            cls, metadict: Mapping[str, meta.Datum]) -> Dict[str, str]:
        iothub_metadata = {}
        for f in metadict:
            if f.startswith('iothub-'):
                v = cls._decode_trigger_metadata_field(
                    metadict, f, python_type=str)
                iothub_metadata[f[len('iothub-'):]] = v
        return iothub_metadata

    @classmethod
    def _extract_iothub_from_system_properties(
            cls, system_properties_string: str) -> Dict[str, str]:
        system_properties = json.loads(system_properties_string)
        return cls._extract_iothub_from_dict(system_properties)

    @classmethod
    def _extract_iothub_from_dict(
            cls, metadict: Dict[str, str]) -> Dict[str, str]:
        iothub_metadata = {}
        for f in metadict:
            if f.startswith('iothub-'):
                iothub_metadata[f[len('iothub-'):]] = metadict[f]
        return iothub_metadata

    @classmethod
    def _is_cardinality_many(cls, trigger_metadata) -> bool:
        return 'SystemPropertiesArray' in trigger_metadata

    @classmethod
    def _is_cardinality_one(cls, trigger_metadata) -> bool:
        return 'SystemProperties' in trigger_metadata



import unittest

from azure.functions import DataType
from azure.functions.decorators.core import BindingDirection
from azure.functions.decorators.openai import (
    AssistantSkillTrigger,
    TextCompletionInput,
    OpenAIModels,
    AssistantQueryInput,
    EmbeddingsInput,
    AssistantCreateOutput,
    SemanticSearchInput,
    EmbeddingsStoreOutput,
    AssistantPostInput,
)


class TestOpenAI(unittest.TestCase):

    def test_assistant_skill_trigger_valid_creation(self):
        trigger = AssistantSkillTrigger(
            name="test",
            function_description="description",
            function_name="test_function_name",
            parameter_description_json="test_json",
            data_type=DataType.UNDEFINED,
            dummy_field="dummy",
        )
        self.assertEqual(trigger.get_binding_name(), "assistantSkillTrigger")
        self.assertEqual(
            trigger.get_dict_repr(),
            {
                "name": "test",
                "functionDescription": "description",
                "functionName": "test_function_name",
                "parameterDescriptionJson": "test_json",
                "dataType": DataType.UNDEFINED,
                "type": "assistantSkillTrigger",
                "dummyField": "dummy",
                "direction": BindingDirection.IN,
            },
        )

    def test_text_completion_input_valid_creation(self):
        input = TextCompletionInput(
            name="test",
            prompt="test_prompt",
            temperature="1",
            max_tokens="1",
            is_reasoning_model=False,
            data_type=DataType.UNDEFINED,
            chat_model=OpenAIModels.DefaultChatModel,
            ai_connection_name="test_connection",
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "textCompletion")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "temperature": "1",
                "maxTokens": "1",
                "type": "textCompletion",
                "dataType": DataType.UNDEFINED,
                "dummyField": "dummy",
                "prompt": "test_prompt",
                "direction": BindingDirection.IN,
                "chatModel": OpenAIModels.DefaultChatModel,
                "isReasoningModel": False,
                "aiConnectionName": "test_connection",
            },
        )

    def test_text_completion_input_with_string_chat_model(self):
        input = TextCompletionInput(
            name="test",
            prompt="test_prompt",
            temperature="1",
            max_tokens="1",
            is_reasoning_model=True,
            data_type=DataType.UNDEFINED,
            chat_model="gpt-4o",
            ai_connection_name="test_connection",
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "textCompletion")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "temperature": "1",
                "maxTokens": "1",
                "type": "textCompletion",
                "dataType": DataType.UNDEFINED,
                "dummyField": "dummy",
                "prompt": "test_prompt",
                "direction": BindingDirection.IN,
                "chatModel": "gpt-4o",
                "isReasoningModel": True,
                "aiConnectionName": "test_connection",
            },
        )

    def test_assistant_query_input_valid_creation(self):
        input = AssistantQueryInput(
            name="test",
            timestamp_utc="timestamp_utc",
            chat_storage_connection_setting="AzureWebJobsStorage",  # noqa: E501
            collection_name="ChatState",
            data_type=DataType.UNDEFINED,
            id="test_id",
            type="assistantQueryInput",
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "assistantQuery")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "timestampUtc": "timestamp_utc",
                "chatStorageConnectionSetting": "AzureWebJobsStorage",  # noqa: E501
                "collectionName": "ChatState",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.IN,
                "type": "assistantQuery",
                "id": "test_id",
                "dummyField": "dummy",
            },
        )

    def test_embeddings_input_valid_creation(self):
        input = EmbeddingsInput(
            name="test",
            data_type=DataType.UNDEFINED,
            input="test_input",
            input_type="test_input_type",
            embeddings_model="test_model",
            max_overlap=1,
            max_chunk_length=1,
            ai_connection_name="test_connection",
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "embeddings")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "type": "embeddings",
                "dataType": DataType.UNDEFINED,
                "input": "test_input",
                "inputType": "test_input_type",
                "embeddingsModel": "test_model",
                "maxOverlap": 1,
                "maxChunkLength": 1,
                "direction": BindingDirection.IN,
                "aiConnectionName": "test_connection",
                "dummyField": "dummy",
            },
        )

    def test_embeddings_input_with_enum_embeddings_model(self):
        input = EmbeddingsInput(
            name="test",
            data_type=DataType.UNDEFINED,
            input="test_input",
            input_type="test_input_type",
            embeddings_model=OpenAIModels.DefaultEmbeddingsModel,
            max_overlap=1,
            max_chunk_length=1,
            ai_connection_name="test_connection",
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "embeddings")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "type": "embeddings",
                "dataType": DataType.UNDEFINED,
                "input": "test_input",
                "inputType": "test_input_type",
                "embeddingsModel": OpenAIModels.DefaultEmbeddingsModel,
                "maxOverlap": 1,
                "maxChunkLength": 1,
                "direction": BindingDirection.IN,
                "aiConnectionName": "test_connection",
                "dummyField": "dummy",
            },
        )

    def test_assistant_create_output_valid_creation(self):
        output = AssistantCreateOutput(
            name="test", data_type=DataType.UNDEFINED
        )
        self.assertEqual(output.get_binding_name(), "assistantCreate")
        self.assertEqual(
            output.get_dict_repr(),
            {
                "name": "test",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.OUT,
                "type": "assistantCreate",
            },
        )

    def test_assistant_post_input_valid_creation(self):
        input = AssistantPostInput(
            name="test",
            id="test_id",
            chat_model="test_model",
            chat_storage_connection_setting="AzureWebJobsStorage",  # noqa: E501
            collection_name="ChatState",
            user_message="test_message",
            temperature="1",
            max_tokens="1",
            is_reasoning_model=False,
            ai_connection_name="test_connection",
            data_type=DataType.UNDEFINED,
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "assistantPost")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "id": "test_id",
                "chatModel": "test_model",
                "chatStorageConnectionSetting": "AzureWebJobsStorage",  # noqa: E501
                "collectionName": "ChatState",
                "userMessage": "test_message",
                "temperature": "1",
                "maxTokens": "1",
                "isReasoningModel": False,
                "aiConnectionName": "test_connection",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.IN,
                "dummyField": "dummy",
                "type": "assistantPost",
            },
        )

    def test_assistant_post_input_with_enum_chat_model(self):
        input = AssistantPostInput(
            name="test",
            id="test_id",
            chat_model=OpenAIModels.DefaultChatModel,
            chat_storage_connection_setting="AzureWebJobsStorage",  # noqa: E501
            collection_name="ChatState",
            user_message="test_message",
            temperature="1",
            max_tokens="1",
            is_reasoning_model=False,
            ai_connection_name="test_connection",
            data_type=DataType.UNDEFINED,
            dummy_field="dummy",
        )
        self.assertEqual(input.get_binding_name(), "assistantPost")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "id": "test_id",
                "chatModel": OpenAIModels.DefaultChatModel,
                "chatStorageConnectionSetting": "AzureWebJobsStorage",  # noqa: E501
                "collectionName": "ChatState",
                "userMessage": "test_message",
                "temperature": "1",
                "maxTokens": "1",
                "isReasoningModel": False,
                "aiConnectionName": "test_connection",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.IN,
                "dummyField": "dummy",
                "type": "assistantPost",
            },
        )

    def test_semantic_search_input_valid_creation(self):
        input = SemanticSearchInput(
            name="test",
            data_type=DataType.UNDEFINED,
            chat_model=OpenAIModels.DefaultChatModel,
            embeddings_model=OpenAIModels.DefaultEmbeddingsModel,  # NoQA
            collection="test_collection",
            search_connection_name="test_connection",
            system_prompt="test_prompt",
            query="test_query",
            max_knowledge_count=1,
            temperature="1",
            max_tokens="1",
            is_reasoning_model=False,
            ai_connection_name="test_connection",
            dummy_field="dummy_field",
        )
        self.assertEqual(input.get_binding_name(), "semanticSearch")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.IN,
                "dummyField": "dummy_field",
                "chatModel": OpenAIModels.DefaultChatModel,
                "embeddingsModel": OpenAIModels.DefaultEmbeddingsModel,  # NoQA
                "type": "semanticSearch",
                "collection": "test_collection",
                "searchConnectionName": "test_connection",
                "systemPrompt": "test_prompt",
                "maxKnowledgeCount": 1,
                "temperature": "1",
                "maxTokens": "1",
                "isReasoningModel": False,
                "aiConnectionName": "test_connection",
                "query": "test_query",
            },
        )

    def test_semantic_search_input_with_string_models(self):
        input = SemanticSearchInput(
            name="test",
            data_type=DataType.UNDEFINED,
            chat_model="gpt-4o",
            embeddings_model="text-embedding-3-large",
            collection="test_collection",
            search_connection_name="test_connection",
            system_prompt="test_prompt",
            query="test_query",
            max_knowledge_count=1,
            temperature="1",
            max_tokens="1",
            is_reasoning_model=True,
            ai_connection_name="test_connection",
            dummy_field="dummy_field",
        )
        self.assertEqual(input.get_binding_name(), "semanticSearch")
        self.assertEqual(
            input.get_dict_repr(),
            {
                "name": "test",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.IN,
                "dummyField": "dummy_field",
                "chatModel": "gpt-4o",
                "embeddingsModel": "text-embedding-3-large",
                "type": "semanticSearch",
                "collection": "test_collection",
                "searchConnectionName": "test_connection",
                "systemPrompt": "test_prompt",
                "maxKnowledgeCount": 1,
                "temperature": "1",
                "maxTokens": "1",
                "isReasoningModel": True,
                "aiConnectionName": "test_connection",
                "query": "test_query",
            },
        )

    def test_embeddings_store_output_valid_creation(self):
        output = EmbeddingsStoreOutput(
            name="test",
            data_type=DataType.UNDEFINED,
            input="test_input",
            input_type="test_input_type",
            store_connection_name="test_connection",
            max_overlap=1,
            max_chunk_length=1,
            collection="test_collection",
            embeddings_model=OpenAIModels.DefaultEmbeddingsModel,  # noqa: E501
            ai_connection_name="test_connection",
            dummy_field="dummy_field",
        )
        self.assertEqual(output.get_binding_name(), "embeddingsStore")
        self.assertEqual(
            output.get_dict_repr(),
            {
                "name": "test",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.OUT,
                "dummyField": "dummy_field",
                "input": "test_input",
                "inputType": "test_input_type",
                "collection": "test_collection",
                "embeddingsModel": OpenAIModels.DefaultEmbeddingsModel,  # noqa: E501
                "storeConnectionName": "test_connection",
                "maxOverlap": 1,
                "maxChunkLength": 1,
                "type": "embeddingsStore",
                "aiConnectionName": "test_connection",
            },
        )

    def test_embeddings_store_output_with_string_embeddings_model(self):
        output = EmbeddingsStoreOutput(
            name="test",
            data_type=DataType.UNDEFINED,
            input="test_input",
            input_type="test_input_type",
            store_connection_name="test_connection",
            max_overlap=1,
            max_chunk_length=1,
            collection="test_collection",
            embeddings_model="text-embedding-3-small",
            ai_connection_name="test_connection",
            dummy_field="dummy_field",
        )
        self.assertEqual(output.get_binding_name(), "embeddingsStore")
        self.assertEqual(
            output.get_dict_repr(),
            {
                "name": "test",
                "dataType": DataType.UNDEFINED,
                "direction": BindingDirection.OUT,
                "dummyField": "dummy_field",
                "input": "test_input",
                "inputType": "test_input_type",
                "collection": "test_collection",
                "embeddingsModel": "text-embedding-3-small",
                "storeConnectionName": "test_connection",
                "maxOverlap": 1,
                "maxChunkLength": 1,
                "type": "embeddingsStore",
                "aiConnectionName": "test_connection",
            },
        )
